from __future__ import annotations

import os
import typing as t

import pytest
import click
from click.testing import CliRunner
from pytest import MonkeyPatch
from PySide6.QtWidgets import QMessageBox, QInputDialog

import clickqt.widgets
from clickqt.core.error import ClickQtError
from tests.testutils import ClickAttrs, clcoancl, raise_, wait_process_Events

clickqt_res: t.Any = None


def callback(p):
    global clickqt_res
    if clickqt_res is None:
        clickqt_res = p
    return p


def prepare_execution(
    monkeypatch: MonkeyPatch, value: t.Any, widget: clickqt.widgets.BaseWidget
) -> tuple[str, t.Optional[str]]:
    if isinstance(widget, clickqt.widgets.MessageBox):
        # Mock the QMessageBox.information-function
        # User clicked on button "Yes" or "No"
        monkeypatch.setattr(
            QMessageBox,
            "information",
            lambda *args: QMessageBox.Yes if value else QMessageBox.No,
        )
    elif (
        isinstance(widget, clickqt.widgets.FileField)
        and value in {"-", "--"}
        and "r" in widget.type.mode
    ):  # "-" -> True; "--" -> False
        monkeypatch.setattr(
            QInputDialog,
            "getMultiLineText",
            lambda *args: (value, value == "-"),
        )  # value, ok

    args: str = ""
    input = None

    if widget.param.multiple:

        def reduce(value) -> str:
            arg_str = "--p="
            for v in value:
                if isinstance(v, list):
                    return reduce(v)
                if str(v) != "":
                    arg_str += str(v) + " "
            return arg_str if arg_str != "--p=" else ""

        for v in value:
            if isinstance(v, list):
                args += reduce(v)
            elif str(v) != "":
                args += f"--p={str(v)} "
    elif isinstance(widget, clickqt.widgets.ConfirmationWidget):
        values = value.split(";")
        args = "--p=" + values[0]
    elif isinstance(widget, clickqt.widgets.MultiWidget):
        args = "--p=" if len(value) > 0 else ""
        for v in value:
            if str(v) != "":
                args += str(v) + " "
            else:  # Don't pass an argument string if any child is empty
                args = ""
                break
    elif not isinstance(widget, clickqt.widgets.MessageBox):
        if isinstance(value, str) and value == "":
            args = ""
        else:
            args = f"--p={str(value)}"
    else:  # widget is MessageBox
        input = "y" if value else "n"

    return (args, input)


@pytest.mark.parametrize(
    ("click_attrs", "value", "error"),
    [
        (ClickAttrs.checkbox(), False, ClickQtError()),
        (ClickAttrs.checkbox(), True, ClickQtError()),
        (ClickAttrs.messagebox(prompt="Test"), False, ClickQtError()),
        (ClickAttrs.messagebox(prompt="Test"), True, ClickQtError()),
        (ClickAttrs.intfield(), 12, ClickQtError()),
        (ClickAttrs.realfield(), -123.2, ClickQtError()),
        (ClickAttrs.intrange(maxval=2, clamp=True), 5, ClickQtError()),
        (ClickAttrs.floatrange(minval=2.5, clamp=True), -1, ClickQtError()),
        (ClickAttrs.textfield(), "test123", ClickQtError()),
        (ClickAttrs.passwordfield(), "abc", ClickQtError()),
        (
            ClickAttrs.confirmation_widget(),
            "test;test",
            ClickQtError(),
        ),  # Testing: split on ';'
        (
            ClickAttrs.combobox(choices=["A", "B", "C"], case_sensitive=False),
            "b",
            ClickQtError(),
        ),
        (ClickAttrs.checkable_combobox(choices=["A", "B", "C"]), [], ClickQtError()),
        (
            ClickAttrs.checkable_combobox(choices=["A", "B", "C"]),
            ["B", "C"],
            ClickQtError(),
        ),
        (ClickAttrs.datetime(formats=["%d-%m-%Y"]), "23-06-2023", ClickQtError()),
        (ClickAttrs.filefield(), ".gitignore", ClickQtError()),
        (ClickAttrs.filefield(type_dict={"mode": "rb"}), "-", ClickQtError()),
        (ClickAttrs.filefield(type_dict={"mode": "w"}), "-", ClickQtError()),
        (ClickAttrs.filefield(type_dict={"mode": "wb"}), "-", ClickQtError()),
        (ClickAttrs.filepathfield(), ".", ClickQtError()),
        (ClickAttrs.tuple_widget(types=()), [], ClickQtError()),
        (
            ClickAttrs.tuple_widget(types=(str, int, float)),
            ("t", 1, -2.0),
            ClickQtError(),
        ),
        (
            ClickAttrs.multi_value_widget(nargs=3, type=float),
            [1.2, "-3.5", -2],
            ClickQtError(),
        ),
        (ClickAttrs.multi_value_widget(nargs=2), ["", ""], ClickQtError()),
        (
            ClickAttrs.nvalue_widget(type=(str, int)),
            [["a", 12], ["c", -1]],
            ClickQtError(),
        ),
        (ClickAttrs.nvalue_widget(type=(str, int)), [], ClickQtError()),
        # Aborted error
        (
            ClickAttrs.messagebox(
                prompt="Test", callback=lambda ctx, param, value: ctx.abort()
            ),
            False,
            ClickQtError(ClickQtError.ErrorType.ABORTED_ERROR),
        ),
        (
            ClickAttrs.filefield(),
            "--",
            ClickQtError(ClickQtError.ErrorType.ABORTED_ERROR),
        ),  # Testing: User wants to input an own message (not from a file) but quits the dialog
        (
            ClickAttrs.nvalue_widget(
                type=(str, int), callback=lambda ctx, param, value: ctx.abort()
            ),
            [["ab", 12]],
            ClickQtError(ClickQtError.ErrorType.ABORTED_ERROR),
        ),
        # Exit error
        (
            ClickAttrs.textfield(callback=lambda ctx, param, value: ctx.exit(1)),
            "abc",
            ClickQtError(ClickQtError.ErrorType.EXIT_ERROR),
        ),
        (
            ClickAttrs.nvalue_widget(
                type=(int, str), callback=lambda ctx, param, value: ctx.exit(1)
            ),
            [[2, "a"]],
            ClickQtError(ClickQtError.ErrorType.EXIT_ERROR),
        ),
        # Converting error (invalid file/path)
        (
            ClickAttrs.filefield(),
            "invalid_file",
            ClickQtError(ClickQtError.ErrorType.CONVERTING_ERROR),
        ),
        (
            ClickAttrs.filepathfield(type_dict={"exists": True}),
            "invalid_path",
            ClickQtError(ClickQtError.ErrorType.CONVERTING_ERROR),
        ),
        (
            ClickAttrs.nvalue_widget(type=(click.types.File(), int)),
            [[".gitignore", 12], ["invalid_file", -1]],
            ClickQtError(ClickQtError.ErrorType.CONVERTING_ERROR),
        ),
        # Processing error (Callback raises an exception)
        (
            ClickAttrs.intfield(
                callback=lambda ctx, param, value: raise_(
                    click.exceptions.BadParameter("...")
                )
            ),
            -3,
            ClickQtError(ClickQtError.ErrorType.PROCESSING_VALUE_ERROR),
        ),
        (
            ClickAttrs.nvalue_widget(
                type=(int, str),
                callback=lambda ctx, param, value: raise_(
                    click.exceptions.BadParameter("...")
                )
                if value[0] != (12, "test")
                else value,
            ),
            [[11, "test"], [231, "abc"]],
            ClickQtError(ClickQtError.ErrorType.PROCESSING_VALUE_ERROR),
        ),
        # Required error (Required param not provided)
        (
            ClickAttrs.textfield(required=True),
            "",
            ClickQtError(ClickQtError.ErrorType.REQUIRED_ERROR),
        ),
        (
            ClickAttrs.checkable_combobox(choices=["A", "B", "C"], required=True),
            [],
            ClickQtError(ClickQtError.ErrorType.REQUIRED_ERROR),
        ),
        (
            ClickAttrs.multi_value_widget(nargs=2, required=True),
            ["", ""],
            ClickQtError(ClickQtError.ErrorType.REQUIRED_ERROR),
        ),
        (
            ClickAttrs.nvalue_widget(type=(str, str), required=True),
            [["", ""]],
            ClickQtError(ClickQtError.ErrorType.REQUIRED_ERROR),
        ),
        (
            ClickAttrs.nvalue_widget(type=(str, str), required=True),
            [],
            ClickQtError(ClickQtError.ErrorType.REQUIRED_ERROR),
        ),
        # With default
        (
            ClickAttrs.textfield(default=""),
            "",
            ClickQtError(),
        ),  # Empty string is accepted
        (
            ClickAttrs.checkable_combobox(choices=["A", "B", "C"], default=["B"]),
            [],
            ClickQtError(),
        ),
        (
            ClickAttrs.multi_value_widget(nargs=2, default=["A", "B"]),
            ["", ""],
            ClickQtError(),
        ),
    ],
)
def test_execution(
    monkeypatch: MonkeyPatch,
    runner: CliRunner,
    click_attrs: dict,
    value: t.Any,
    error: ClickQtError,
):
    global clickqt_res

    param = click.Option(param_decls=["--p"], **click_attrs)
    cli = click.Command("cli", params=[param], callback=callback)

    control = clickqt.qtgui_from_click(cli)
    widget = control.widget_registry[cli.name][param.name]

    if (
        isinstance(widget, clickqt.widgets.FileField)
        and value == "--"
        and "r" in widget.type.mode
    ):
        widget.set_value("-")
    elif isinstance(widget, clickqt.widgets.ConfirmationWidget):
        values = value.split(";")
        widget.field.set_value(values[0])
        widget.confirmation_field.set_value(values[1])
    elif value is not None:
        widget.set_value(value)

    args, inputs = prepare_execution(monkeypatch, value, widget)
    standalone_mode = False
    if error.type == ClickQtError.ErrorType.EXIT_ERROR:  #  See click/core.py#1082
        standalone_mode = True
    click_res = runner.invoke(cli, args, inputs, standalone_mode=standalone_mode)
    val, err = widget.get_value()

    if (
        isinstance(widget, clickqt.widgets.FileField)
        and "r" in widget.type.mode
        and widget.get_widget_value() == "-"
    ):
        assert callable(val)
        val, err = val()

    # First compare the value from 'widget.get_value()' with the click result
    # then the clickqt result (run_button clicked) with the click result
    for i in range(2):
        if not isinstance(widget, clickqt.widgets.FileField):
            assert val == click_res.return_value
        else:  # IOWrapper-objects can't be compared
            closest_common_ancestor_class = clcoancl(
                type(val), type(click_res.return_value)
            )
            assert closest_common_ancestor_class not in {None, object}

        if i == 0:
            assert err.type == error.type
            if error.type == ClickQtError.ErrorType.ABORTED_ERROR:
                if (
                    isinstance(widget, clickqt.widgets.FileField)
                    and value == "--"
                    and "r" in widget.type.mode
                ):
                    assert (
                        isinstance(click_res.exception, click.exceptions.BadParameter)
                        and "'--': No such file or directory"
                        in click_res.exception.message
                    )
                else:
                    assert isinstance(click_res.exception, click.exceptions.Abort)
            elif error.type == ClickQtError.ErrorType.EXIT_ERROR:
                assert isinstance(
                    click_res.exception, SystemExit
                )  # Not click.exceptions.Exit, see click/core.py#1082
            elif error.type == ClickQtError.ErrorType.REQUIRED_ERROR:
                assert isinstance(
                    click_res.exception, click.exceptions.MissingParameter
                )
            elif error.type in {
                ClickQtError.ErrorType.CONVERTING_ERROR,
                ClickQtError.ErrorType.PROCESSING_VALUE_ERROR,
            }:
                assert isinstance(click_res.exception, click.exceptions.BadParameter)
            else:
                assert click_res.exception is None

            clickqt_res = None  # Reset the stored click result
            control.gui.run_button.click()
            wait_process_Events(1)  # Wait for worker thread to finish the execution
            val = clickqt_res


@pytest.mark.parametrize(
    ("click_attrs", "value1", "value2", "error"),
    [
        (
            ClickAttrs.confirmation_widget(),
            "a",
            "b",
            ClickQtError(ClickQtError.ErrorType.CONFIRMATION_INPUT_NOT_EQUAL_ERROR),
        ),
        (
            ClickAttrs.confirmation_widget(required=True),
            "",
            "b",
            ClickQtError(ClickQtError.ErrorType.REQUIRED_ERROR),
        ),  # First check for required, then value equality
    ],
)
def test_execution_confirmation_widget_fail(
    click_attrs: dict, value1: str, value2: str, error: ClickQtError
):
    param = click.Option(param_decls=["--p"], **click_attrs)
    cli = click.Command("cli", params=[param])

    control = clickqt.qtgui_from_click(cli)
    widget: clickqt.widgets.ConfirmationWidget = control.widget_registry[cli.name][
        param.name
    ]

    widget.field.set_value(value1)
    widget.confirmation_field.set_value(value2)

    val, err = widget.get_value()

    assert val is None and err.type == error.type


@pytest.mark.parametrize(
    ("click_attrs", "value", "envvar_values"),
    [
        (ClickAttrs.nvalue_widget(), [], []),
        (ClickAttrs.nvalue_widget(), [], ["test1", "test2"]),
        (ClickAttrs.nvalue_widget(), ["a", "b"], []),
        (ClickAttrs.nvalue_widget(), ["a", "b"], ["test1", "test2"]),
        (ClickAttrs.nvalue_widget(default=["x", "y"]), [], []),
        (ClickAttrs.nvalue_widget(default=["x", "y"]), [], ["test1", "test2"]),
        (ClickAttrs.nvalue_widget(default=["x", "y"]), ["a", "b"], []),
        (ClickAttrs.nvalue_widget(default=["x", "y"]), ["a", "b"], ["test1", "test2"]),
        # Children are empty
        (ClickAttrs.nvalue_widget(default=["x", "y"]), ["", ""], []),
        (ClickAttrs.nvalue_widget(default=[]), ["", ""], []),
    ],
)
def test_execution_nvalue_widget(
    runner: CliRunner,
    click_attrs: dict,
    value: t.Sequence[str],
    envvar_values: t.Sequence[str],
):
    global clickqt_res
    os.environ["TEST_CLICKQT_ENVVAR"] = os.path.pathsep.join(envvar_values)

    param = click.Option(
        param_decls=["--p"], envvar="TEST_CLICKQT_ENVVAR", **click_attrs
    )
    cli = click.Command("cli", params=[param], callback=callback)

    control = clickqt.qtgui_from_click(cli)
    widget: clickqt.widgets.NValueWidget = control.widget_registry[cli.name][param.name]

    widget.set_value(value)

    val, err = widget.get_value()
    args, input = prepare_execution(monkeypatch=None, value=value, widget=widget)
    click_res = runner.invoke(cli, args, input, standalone_mode=False)
    for i in range(2):
        assert val == click_res.return_value

        if i == 0:
            assert err.type == ClickQtError.ErrorType.NO_ERROR

            clickqt_res = None  # Reset the stored click result
            control.gui.run_button.click()
            wait_process_Events(1)  # Wait for worker thread to finish the execution
            val = clickqt_res


def test_execution_context():
    clickqt_res: list = []

    @click.group()
    @click.pass_context
    def cli(ctx):
        nonlocal clickqt_res
        ctx.obj = "test1"
        clickqt_res.append(ctx)

    @cli.command()
    @click.pass_obj
    def test(obj):
        nonlocal clickqt_res
        clickqt_res.append(obj)

    control = clickqt.qtgui_from_click(cli)
    control.gui.run_button.click()

    wait_process_Events(100)  # Wait for worker thread to finish the execution

    assert (
        len(clickqt_res) == 2
        and isinstance(clickqt_res[0], click.Context)
        and clickqt_res[1] == "test1"
    )


def test_execution_expose_value_kwargs():
    clickqt_res: dict = None

    def f(p1, **kwargs):
        nonlocal clickqt_res
        clickqt_res = kwargs

    cli = click.Command(
        "cli",
        params=[
            click.Option(["--p1"], default="a"),
            click.Option(["--p2"], default="b", expose_value=False),
            click.Option(["--p3"], default="c"),
        ],
        callback=f,
    )

    control = clickqt.qtgui_from_click(cli)
    control.gui.run_button.click()

    wait_process_Events(1)  # Wait for worker thread to finish the execution

    assert len(clickqt_res.values()) == 1
    assert clickqt_res.get("p3") == "c"



from __future__ import annotations

import sys
import typing as t

import click
from PySide6.QtCore import Signal, QObject, Slot


class CommandExecutor(QObject):
    """Worker which executes the received tasks/callbacks"""

    finished: Signal = Signal()
    # Internal Qt-signal emitted when :func:`~clickqt.core.commandexecutor.CommandExecutor.run`-Slot has finished

    @Slot(list, click.Context)
    def run(
        self, tasks: t.Iterable[t.Callable], ctx: click.Context
    ):  # pragma: no cover; Tested in test_execution.py
        """Pushes the current context on the click internal stack and executes the received tasks.
        When the execution is done, the finished signal will be emitted

        :param tasks: The callbacks to execute
        :param ctx: The current context which should be pushed on the click internal stack
        """

        # Push context of selected command, needed for @click.pass_context and @click.pass_obj
        click.globals.push_context(ctx)

        for task in tasks:
            try:
                task()
            except SystemExit as e:
                print(f"SystemExit-Exception, return code: {e.code}", file=sys.stderr)
            except Exception as e: # pylint: disable=broad-exception-caught
                print(f"Exception: {str(e)}", file=sys.stderr)

        self.finished.emit()



from __future__ import annotations

import typing as t

import click
from PySide6.QtWidgets import QGroupBox, QVBoxLayout

from clickqt.widgets.basewidget import BaseWidget, MultiWidget


class MultiValueWidget(MultiWidget):
    """Represents a multi value click option with a fixed number of types (**param**.nargs >=2).
    The child widgets are set according to :func:`~clickqt.widgets.basewidget.MultiWidget.init`.

    :param otype: The type which specifies the clickqt widget type. This type may be different compared to **param**.type when dealing with click.types.CompositeParamType-objects
    :param param: The parameter from which **otype** came from
    :param widgetsource: A reference to :func:`~clickqt.core.gui.GUI.create_widget`
    :param parent: The parent BaseWidget of **otype**, defaults to None. Needed for :class:`~clickqt.widgets.basewidget.MultiWidget`-widgets
    :param kwargs: Additionally parameters ('com', 'label') needed for
                    :class:`~clickqt.widgets.basewidget.MultiWidget`- / :class:`~clickqt.widgets.confirmationwidget.ConfirmationWidget`-widgets
    """

    widget_type = QGroupBox  #: The Qt-type of this widget.

    def __init__(
        self,
        otype: click.ParamType,
        param: click.Parameter,
        widgetsource: t.Callable[[t.Any], BaseWidget],
        parent: t.Optional[BaseWidget] = None,
        **kwargs,
    ):
        super().__init__(otype, param, parent=parent, **kwargs)

        assert (
            param.nargs >= 2
        ), f"'param.nargs' should be >= 2, but is '{param.nargs}'."

        self.widget.setLayout(QVBoxLayout())

        # Add param.nargs widgets of type otype
        for _ in range(param.nargs):
            nargs = param.nargs
            param.nargs = 1  # Stop recursion
            bw: BaseWidget = widgetsource(
                otype, param, widgetsource=widgetsource, parent=self, **kwargs
            )
            param.nargs = nargs  # click needs the right value for a correct conversion
            bw.layout.removeWidget(bw.label)
            bw.label.deleteLater()
            self.widget.layout().addWidget(bw.container)
            self.children.append(bw)

        self.init()



from __future__ import annotations

import typing as t
import sys
from functools import reduce
import re
import inspect
import click
from PySide6.QtWidgets import (
    QWidget,
    QFrame,
    QVBoxLayout,
    QTabWidget,
    QScrollArea,
    QApplication,
    QSizePolicy,
    QLabel,
)
from PySide6.QtCore import QThread, QObject, Signal, Slot, Qt
from PySide6.QtGui import QPalette, QClipboard

from clickqt.core.gui import GUI
from clickqt.core.commandexecutor import CommandExecutor
from clickqt.core.error import ClickQtError
from clickqt.widgets.combobox import CheckableComboBox, ComboBox
from clickqt.widgets.basewidget import BaseWidget
from clickqt.widgets.messagebox import MessageBox
from clickqt.widgets.filefield import FileField
from clickqt.core.utils import is_nested_list, is_file_path


class Control(QObject):
    """Regulates the creation of the GUI with their widgets according to clicks parameter types and causes the execution/abortion of a selected command.

    :param cmd: The callback function from which a GUI should be created
    """

    #: Internal Qt-signal, which will be emitted when the :func:`~clickqt.core.control.Control.start_execution`-Slot was triggered and executed successfully.
    requestExecution: Signal = Signal(list, click.Context)  # Generics do not work here

    def __init__(
        self, cmd: click.Command, is_ep: bool = True, ep_or_path: str = "test"
    ):
        """Initializing the GUI object and the registries together with the differentiation of a group command and a simple command."""

        super().__init__()

        self.gui = GUI()
        self.cmd = cmd

        self.is_ep = is_ep
        self.ep_or_path = ep_or_path

        # Create a worker in another thread when the user clicks the run button
        # Don't destroy a thread when no command is running and the user closes the application
        # Otherwise "QThread: Destroyed while thread is still running" would be appear
        self.worker_thread: QThread = None
        self.worker: CommandExecutor = None

        # Connect GUI buttons with slots
        self.gui.run_button.clicked.connect(self.start_execution)
        self.gui.stop_button.clicked.connect(self.stop_execution)
        self.gui.copy_button.clicked.connect(self.construct_command_string)

        # Groups-Command-name concatinated with ":" to command-option-names to BaseWidget
        self.widget_registry: dict[str, dict[str, BaseWidget]] = {}
        self.command_registry: dict[str, dict[str, tuple[int, t.Callable]]] = {}

        # Add all widgets
        self.parse(self.gui.widgets_container, cmd, cmd.name)

        self.gui.construct()

    def __call__(self):
        """Shows the GUI according to :func:`~clickqt.core.gui.GUI.__call__` of :class:`~clickqt.core.gui.GUI`."""

        self.gui()

    def set_ep_or_path(self, ep_or_path):
        self.ep_or_path = ep_or_path

    def set_is_ep(self, is_ep):
        self.is_ep = is_ep

    def parameter_to_widget(
        self, command: click.Command, groups_command_name: str, param: click.Parameter
    ) -> QWidget:
        """Creates a clickqt widget according to :func:`~clickqt.core.gui.GUI.create_widget` and returns the container of the widget (label-element + Qt-widget).

        :param command: The click command of the provided **param**
        :param groups_command_name: The hierarchy of the **command** as string whereby the names of the components are
                                    concatenated according to :func:`~clickqt.core.control.Control.concat`
        :param param: The click parameter whose type a clickqt widget should be created from

        :return: The container of the created widget (label-element + Qt-widget)
        """

        assert param.name, "No parameter name specified"
        assert self.widget_registry[groups_command_name].get(param.name) is None

        widget = self.gui.create_widget(
            param.type, param, widgetsource=self.gui.create_widget, com=command
        )
        self.widget_registry[groups_command_name][param.name] = widget
        self.command_registry[groups_command_name][param.name] = (
            param.nargs,
            type(param.type).__name__,
        )

        return widget.container

    def concat(self, a: str, b: str) -> str:  # pylint: disable=no-self-use
        """Concatenates the strings a and b with ':' and returns the result."""

        return a + ":" + b

    def parse(
        self,
        tab_widget: QWidget,
        cmd: click.Command,
        group_name: str,
        group_names_concatenated: str = "",
    ):
        if isinstance(cmd, click.Group):
            child_tabs: QWidget = None
            concat_group_names = (
                self.concat(group_names_concatenated, group_name)
                if group_names_concatenated
                else group_name
            )
            if len(cmd.params) > 0:
                child_tabs = QWidget()
                child_tabs.setLayout(QVBoxLayout())
                group_params = self.parse_cmd(cmd, concat_group_names)
                group_params.widget().layout().setContentsMargins(0, 0, 0, 0)
                group_params.setSizePolicy(
                    QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed
                )  # Group params don't have to be resizable
                child_tabs.layout().addWidget(group_params)
                child_tabs.layout().addWidget(
                    self.parse_cmd_group(cmd, concat_group_names)
                )
            else:
                child_tabs = self.parse_cmd_group(cmd, concat_group_names)

            child_tabs.setAutoFillBackground(True)
            child_tabs.setBackgroundRole(
                QPalette.ColorRole.Window
            )  # Remove white spacing between widgets

            if tab_widget == self.gui.widgets_container:
                self.gui.widgets_container = child_tabs
            else:
                tab_widget.addTab(child_tabs, group_name)
        elif tab_widget == self.gui.widgets_container:
            self.gui.widgets_container = self.parse_cmd(cmd, cmd.name)
        else:
            tab_widget.addTab(
                self.parse_cmd(
                    cmd,
                    self.concat(group_names_concatenated, cmd.name)
                    if group_names_concatenated
                    else cmd.name,
                ),
                group_name,
            )

    def parse_cmd_group(
        self, cmdgroup: click.Group, group_names_concatenated: str
    ) -> QTabWidget:
        """Creates for every group in **cmdgroup** a QTabWidget instance and adds every command in **cmdgroup** as a tab to it.
        The creation of the content of every tab is realized by calling :func:`~clickqt.core.control.Control.parse_cmd`.
        To realize command hierachies, this method is called recursively.

        :param cmdgroup: The group from which a QTabWidget with content should be created
        :param group_names_concatenated: The hierarchy of **cmdgroup** as string whereby the names of the components are
                                         concatenated according to :func:`~clickqt.core.control.Control.concat`

        :returns: A Qt-GUI representation in a QTabWidget of **cmdgroup**
        """

        group_tab_widget = QTabWidget()
        for group_name, group_cmd in cmdgroup.commands.items():
            self.parse(
                group_tab_widget, group_cmd, group_name, group_names_concatenated
            )

        return group_tab_widget

    def parse_cmd(self, cmd: click.Command, groups_command_name: str) -> QScrollArea:
        """Creates for every click parameter in **cmd** a clickqt widget and returns them stored in a QScrollArea.
        The widgets are divided into a "Required arguments" and "Optional arguments" part.

        :param cmd: The command from which a QTabWidget with content should be created
        :param groups_command_name: The hierarchy of **cmd** as string whereby the names of the components are
                                    concatenated according to :func:`~clickqt.core.control.Control.concat`

        :returns: The created clickqt widgets stored in a QScrollArea
        """

        cmdbox = QWidget()
        cmdbox.setLayout(QVBoxLayout())
        cmdbox.layout().setAlignment(Qt.AlignmentFlag.AlignTop)

        required_optional_box: list[QWidget] = []

        for i in range(2):
            box = QWidget()
            box.setLayout(QVBoxLayout())
            box_label = QLabel(
                text=f"<b>{'Required arguments' if i == 0 else 'Optional arguments'}</b>"
            )
            box_label.setTextFormat(Qt.TextFormat.RichText)  # Bold text
            box.layout().addWidget(box_label)
            line = QFrame()
            line.setFrameShape(QFrame.Shape.HLine)
            box.layout().addWidget(line)
            box.layout().setAlignment(Qt.AlignmentFlag.AlignTop)

            required_optional_box.append(box)

        INITIAL_CHILD_WIDGETS = len(
            required_optional_box[0].children()
        )  # layout, label, line

        assert (
            self.widget_registry.get(groups_command_name) is None
        ), f"Not a unique group_command_name_concat ({groups_command_name})"

        self.widget_registry[groups_command_name] = {}
        self.command_registry[groups_command_name] = {}

        # parameter name to flag values
        feature_switches: dict[str, list[click.Parameter]] = {}

        for param in cmd.params:
            if isinstance(param, click.core.Parameter):
                if (
                    hasattr(param, "is_flag")
                    and param.is_flag
                    and hasattr(param, "flag_value")
                    and isinstance(param.flag_value, str)
                    and param.flag_value
                ):  # clicks feature switches
                    if feature_switches.get(param.name) is None:
                        feature_switches[param.name] = []
                    feature_switches[param.name].append(param)
                else:
                    required_optional_box[
                        0
                        if (param.required or isinstance(param, click.Argument))
                        and param.default is None
                        else 1
                    ].layout().addWidget(
                        self.parameter_to_widget(cmd, groups_command_name, param)
                    )

        # Create for every feature switch a ComboBox
        for param_name, switch_names in feature_switches.items():
            choice = click.Option(
                [f"--{param_name}"],
                type=click.Choice([x.flag_value for x in switch_names]),
                required=reduce(lambda x, y: x | y.required, switch_names, False),
            )
            default = next(
                (x.flag_value for x in switch_names if x.default),
                switch_names[0].flag_value,
            )  # First param with default==True is the default
            required_optional_box[0 if choice.required else 1].layout().addWidget(
                self.parameter_to_widget(cmd, groups_command_name, choice)
            )
            self.widget_registry[groups_command_name][param_name].set_value(default)

        for box in required_optional_box:
            if len(box.children()) > INITIAL_CHILD_WIDGETS:
                cmdbox.layout().addWidget(box)

        cmd_tab_widget = QScrollArea()
        cmd_tab_widget.setFrameShape(QFrame.Shape.NoFrame)  # Remove black border
        cmd_tab_widget.setBackgroundRole(QPalette.ColorRole.Window)
        cmd_tab_widget.setWidgetResizable(True)  # Widgets should use the whole area
        cmd_tab_widget.setWidget(cmdbox)

        return cmd_tab_widget

    def check_error(self, err: ClickQtError) -> bool:
        """Checks whether **err** contains an error and prints on error case the message of it to sys.stderr.

        :return: True, if **err** contains an error, False otherwise"""

        if err.type != ClickQtError.ErrorType.NO_ERROR:
            if message := err.message():  # Don't print on context exit
                print(message, file=sys.stderr)
            return True

        return False

    def current_command_hierarchy(
        self, tab_widget: QWidget, cmd: click.Command
    ) -> list[click.Command]:
        """Returns the hierarchy of the command of the selected tab as list whereby the order of the list is from root command
        to the selected command.

        :param tab_widget: The currend widget of the root-QTabWidget
        :param cmd: The click command provided to :func:`~clickqt.core.control.Control`

        :return: The hierarchy of the command of the selected tab as ordered list (root command to selected command)
        """

        if isinstance(cmd, click.Group):
            if len(cmd.params) > 0:  # Group has params
                tab_widget = tab_widget.findChild(QTabWidget)

            assert isinstance(tab_widget, QTabWidget)

            command = cmd.get_command(
                ctx=None, cmd_name=tab_widget.tabText(tab_widget.currentIndex())
            )

            return [cmd] + self.current_command_hierarchy(
                tab_widget.currentWidget(), command
            )

        return [cmd]

    def get_option_names(self, cmd):
        """Returns an array of all the parameters used for the current command togeter with their properties."""
        option_names = []
        for param in cmd.params:
            if isinstance(param, click.Option):
                long_forms = [opt for opt in param.opts if opt.startswith("--")]
                longest_long_form = max(long_forms, key=len) if long_forms else None
                short_forms = [opt for opt in param.opts if opt.startswith("-")]
                short_forms = max(short_forms, key=len) if short_forms else None
                if longest_long_form:
                    option_names.append(
                        (
                            longest_long_form,
                            param.type,
                            param.multiple,
                            param.nargs,
                            param.confirmation_prompt,
                        )
                    )
                else:
                    option_names.append(
                        (
                            short_forms,
                            param.type,
                            param.multiple,
                            param.nargs,
                            param.confirmation_prompt,
                        )
                    )
            elif isinstance(param, click.Argument):
                option_names.append(("Argument", param.type))
        return option_names

    def get_params(self, selected_command_name: str, args):
        """Returns an array of strings that are used for the output field."""
        params = [k for k, v in self.widget_registry[selected_command_name].items()]
        if "yes" in params:
            params.remove("yes")
        command_help = self.command_registry.get(selected_command_name)
        tuples_array = list(command_help.values())
        for i, param in enumerate(args):
            params[i] = "--" + param + f": {tuples_array[i]}: " + f"{args[param]}"
        return params

    def clean_command_string(self, word, text):
        """Returns a string without any special characters using regex."""
        text = re.sub(r"\b{}\b".format(re.escape(word)), "", text)
        text = re.sub(r"[^a-zA-Z0-9 .-]", " ", text)
        return text

    def command_to_string(self, hierarchy_selected_command_name: str):
        """Returns the current command name."""
        hierarchy_selected_command_name = self.clean_command_string(
            self.cmd.name, hierarchy_selected_command_name
        )
        return self.ep_or_path + " " + hierarchy_selected_command_name

    def command_to_string_to_copy(self, hierarchy_selected_name: str, selected_command):
        """Returns a string representing the click command if one actually would actually execute it in the shell."""
        parameter_list = self.get_option_names(selected_command)
        parameter_list = [param for param in parameter_list if param[0] != "--yes"]
        widgets = self.widget_registry[hierarchy_selected_name]
        widget_keys = list(widgets.keys())
        if "yes" in widgets:
            widgets.pop("yes")
        widget_values = []
        for widget in widgets:
            if widget != "yes":
                widget_values.append(widgets[widget].get_widget_value())
        parameter_strings = []
        for i, param in enumerate(parameter_list):
            if param[0] == "Argument":
                parameter_strings.append(str(widget_values[i]))
                continue
            if (not isinstance(widget_values[i], list)) and param[2] is not True:
                widget_value = str(widget_values[i])
                if (isinstance(param[1], click.Choice) and param[4] is True) or (
                    isinstance(param[1], click.Choice)
                ):
                    if is_file_path(widget_value):
                        parameter_strings.append(param[0] + "=" + widget_value)
                    else:
                        parameter_strings.append(
                            param[0]
                            + "="
                            + re.sub(r"[^a-zA-Z0-9 .-]", " ", widget_value)
                        )
                else:
                    if is_file_path(widget_value):
                        parameter_strings.append(param[0] + " " + widget_value)
                    else:
                        parameter_strings.append(
                            param[0]
                            + " "
                            + re.sub(r"[^a-zA-Z0-9 .-]", " ", widget_value)
                        )
            else:
                if is_nested_list(widget_values[i]):
                    depth = len(widget_values[i])
                    for j in range(depth):
                        widget_value = str(widget_values[i][j])
                        if not is_file_path(widget_value):
                            parameter_strings.append(
                                param[0]
                                + " "
                                + re.sub(r"[^a-zA-Z0-9 .-]", " ", widget_value)
                            )
                        else:
                            parameter_strings.append(param[0] + " " + widget_value)
                else:
                    length = len(widget_values[i])
                    if param[2] is not True:
                        parameter_strings.append(param[0])
                        for j in range(length):
                            widget_value = str(widget_values[i][j])
                            if not is_file_path(widget_value):
                                parameter_strings.append(
                                    " " + re.sub(r"[^a-zA-Z0-9 .-]", " ", widget_value)
                                )
                            else:
                                parameter_strings.append(" " + widget_value)
                    else:
                        if isinstance(widgets[widget_keys[i]], CheckableComboBox):
                            for j in range(length):
                                widget_value = str(widget_values[i][j])
                                if not is_file_path(widget_value):
                                    parameter_strings.append(
                                        param[0]
                                        + "="
                                        + re.sub(r"[^a-zA-Z0-9 .-]", " ", widget_value)
                                    )
                                else:
                                    parameter_strings.append(
                                        param[0] + "=" + widget_value
                                    )
                        else:
                            for j in range(length):
                                widget_value = str(widget_values[i][j])
                                if not is_file_path(widget_value):
                                    parameter_strings.append(
                                        param[0]
                                        + " "
                                        + re.sub(r"[^a-zA-Z0-9 .-]", " ", widget_value)
                                    )
                                else:
                                    parameter_strings.append(
                                        param[0] + " " + widget_value
                                    )
        message = hierarchy_selected_name + " " + " ".join(parameter_strings)
        message = re.sub(r"\b{}\b".format(re.escape(self.cmd.name)), "", message)
        message = message.replace(":", " ")
        if not self.is_ep:
            message = "python " + self.ep_or_path + " " + message
        else:
            message = self.ep_or_path + " " + message
        return message

    def function_call_formatter(
        self, hierarchy_selected_command_name: str, selected_command_name: str, args
    ):
        params = self.get_params(hierarchy_selected_command_name, args)
        message = f"{selected_command_name} \n"
        parameter_message = "Current Command parameters: \n" + "\n".join(params)
        return message + parameter_message

    @Slot()
    def stop_execution(self):
        """Qt-Slot, which stops the execution of the command(-hierarchy) which is currently running."""

        print("Execution stopped!", file=sys.stderr)
        self.worker_thread.terminate()
        self.execution_finished()

    @Slot()
    def execution_finished(self):
        """Qt-Slot, which deletes the internal worker-object and resets the buttons of the GUI.
        This slot is automatically executed when the execution of a command has finished.
        """

        self.worker_thread.deleteLater()
        self.worker.deleteLater()

        self.worker_thread = None
        self.worker = None

        self.gui.run_button.setEnabled(True)
        self.gui.stop_button.setEnabled(False)

    @Slot()
    def start_execution(self):
        """Qt-Slot, which validates the selected command hierarchy and causes (on success) their execution in another thread by
        emitting the :func:`~clickqt.core.control.Control.requestExecution`-Signal. Widgets that will show a dialog will be validated at last.
        This slot is automatically executed when the user clicks on the 'Run'-button.
        """

        hierarchy_selected_command = self.current_command_hierarchy(
            self.gui.widgets_container, self.cmd
        )

        def run_command(
            command: click.Command, hierarchy_command: str
        ) -> t.Optional[t.Callable]:
            kwargs: dict[str, t.Any] = {}
            has_error = False
            dialog_widgets: list[BaseWidget] = []  # widgets that will show a dialog

            if (
                self.widget_registry.get(hierarchy_command) is not None
            ):  # Groups with no options are not in the dict
                # Check the values of all non dialog widgets for errors
                for option_name, widget in self.widget_registry[
                    hierarchy_command
                ].items():
                    if isinstance(widget, MessageBox):
                        dialog_widgets.append(
                            widget
                        )  # MessageBox widgets should be shown at last
                    elif (
                        isinstance(widget, FileField)
                        and "r" in widget.type.mode
                        and widget.get_widget_value() == "-"
                    ):
                        dialog_widgets.insert(
                            0, widget
                        )  # FileField widgets with input dialog should be shown at last, but before MessageBox widgets
                    else:
                        widget_value, err = widget.get_value()
                        has_error |= self.check_error(err)

                        if widget.param.expose_value:
                            kwargs[option_name] = widget_value

                if has_error:
                    return None

                # Now check the values of all dialog widgets for errors
                for widget in dialog_widgets:
                    widget_value, err = widget.get_value()
                    if isinstance(widget, FileField):
                        assert callable(widget_value)
                        widget_value, err = widget_value()

                    if self.check_error(err):
                        return None

                    if widget.param.expose_value:
                        kwargs[widget.param.name] = widget_value

            if len(callback_args := inspect.getfullargspec(command.callback).args) > 0:
                args: list[t.Any] = []
                for ca in callback_args:  # Bring the args in the correct order
                    args.append(
                        kwargs.pop(ca)
                    )  # Remove explicitly mentioned args from kwargs

                    print(
                        f"For command details, please call '{self.command_to_string(hierarchy_command)} --help'"
                    )
                    print(self.command_to_string_to_copy(hierarchy_command, command))
                    print(
                        f"Current Command: {self.function_call_formatter(hierarchy_command, command, kwargs)} \n"
                        + "Output:"
                    )
                    return lambda: command.callback(*args, **kwargs)
            else:
                return lambda: command.callback(
                    **kwargs
                )  # pylint: disable=unnecessary-lambda

        callables: list[t.Callable] = []
        for i, command in enumerate(hierarchy_selected_command, 1):
            if (
                c := run_command(
                    command,
                    reduce(
                        self.concat, [g.name for g in hierarchy_selected_command[:i]]
                    ),
                )
            ) is not None:
                callables.append(c)

        if len(callables) == len(hierarchy_selected_command):
            self.gui.run_button.setEnabled(False)
            self.gui.stop_button.setEnabled(True)

            self.worker_thread = QThread()
            self.worker_thread.start()
            self.worker = CommandExecutor()
            self.worker.moveToThread(self.worker_thread)
            self.worker.finished.connect(self.worker_thread.quit)
            self.worker.finished.connect(self.execution_finished)
            self.requestExecution.connect(self.worker.run)

            self.requestExecution.emit(
                callables, click.Context(hierarchy_selected_command[-1])
            )

    def construct_command_string(self):
        """
        This function is responsible
        """
        hierarchy_selected_command = self.current_command_hierarchy(
            self.gui.widgets_container, self.cmd
        )
        selected_command = hierarchy_selected_command[-1]
        hierarchy_selected_command_name = reduce(
            self.concat, [g.name for g in hierarchy_selected_command]
        )

        kwargs: dict[str, t.Any] = {}
        has_error = False
        unused_options: list[BaseWidget] = []  # parameters with expose_value==False

        # Check all values for errors
        for option_name, widget in self.widget_registry[
            hierarchy_selected_command_name
        ].items():
            param: click.Parameter = next(
                (x for x in selected_command.params if x.name == option_name)
            )
            if param.expose_value:
                widget_value, err = widget.get_value()
                has_error |= self.check_error(err)

                kwargs[option_name] = widget_value
            else:  # Verify it when all options are valid
                unused_options.append(widget)

        if has_error:
            return

        # Replace the callables with their values and check for errors
        for option_name, value in kwargs.items():
            if callable(value):
                kwargs[option_name], err = value()
                has_error |= self.check_error(err)

        if has_error:
            return

        # Parameters with expose_value==False
        for widget in unused_options:
            widget_value, err = widget.get_value()
            has_error |= self.check_error(err)
            if callable(widget_value):
                _, err = widget_value()
                has_error |= self.check_error(err)

        if has_error:
            return

        message = self.command_to_string_to_copy(
            hierarchy_selected_command_name, selected_command
        )
        clip_board = QApplication.clipboard()
        clip_board.setText(message, QClipboard.Clipboard)



from __future__ import annotations

from enum import IntEnum

try:
    from enum_tools.documentation import document_enum
except ImportError:  # pragma: no cover
    document_enum = lambda x: x # pylint: disable=unnecessary-lambda-assignment


class ClickQtError:
    """Shows whether an error occured during value validation of the widgets.

    :param type: The type of the error, defaults to ErrorType.NO_ERROR
    :param trigger: The source where the error occured, defaults to ''
    :param click_error_message: The error message which was created by click, defaults to ''
    """

    @document_enum
    class ErrorType(IntEnum):
        """Specifies the possible error types."""

        NO_ERROR = 0  # doc: No error occured.
        CONFIRMATION_INPUT_NOT_EQUAL_ERROR = 1  # doc: The values of the elements of a ConfirmationWidget do not match.
        ABORTED_ERROR = 2  # doc: A clickqt dialog / the click.Context-object of the command was aborted.
        PROCESSING_VALUE_ERROR = 3  # doc: An error was triggered in a (user-defined) callback of a click.Parameter.
        CONVERTING_ERROR = 4  # doc: A value could not be converted into a click.ParamType.
        REQUIRED_ERROR = 5  # doc: The value of a required option is missing.
        EXIT_ERROR = 6  # doc: exit() was called on click.Context-object of the command.

    def __init__(
        self,
        type: "ErrorType" = ErrorType.NO_ERROR,
        trigger: str = "",
        click_error_message: str = "",
    ):
        self.type = type
        self.trigger = trigger
        self.click_error_message = click_error_message

    def message(self) -> str:
        """Returns the message of the error as string.
        If there was no error, an empty string will be returned.
        """

        if self.type.value == ClickQtError.ErrorType.NO_ERROR:
            return ""
        if (
            self.type.value == ClickQtError.ErrorType.CONFIRMATION_INPUT_NOT_EQUAL_ERROR
        ):
            return f"Confirmation input ({self.trigger}) is not equal"
        if self.type.value == ClickQtError.ErrorType.ABORTED_ERROR:
            return "Aborted!"
        if self.type.value == ClickQtError.ErrorType.PROCESSING_VALUE_ERROR:
            return (
                f"Processing value error ({self.trigger}): {self.click_error_message}"
            )
        if self.type.value == ClickQtError.ErrorType.CONVERTING_ERROR:
            return f"Converting error ({self.trigger}): {self.click_error_message}"
        if self.type.value == ClickQtError.ErrorType.REQUIRED_ERROR:
            return f"Required error ({self.trigger}): {self.click_error_message} is empty"  # Argument/Option
        if self.type.value == ClickQtError.ErrorType.EXIT_ERROR:
            return ""  # Don't print an error (click behaviour)

        raise NotImplementedError(
            f"Message for this error ({self.type.value}) not implemented yet"
        )



"""
Configures the test setup
"""

import re
import typing as t

import pytest

from click.testing import CliRunner


@pytest.fixture(scope="function")
def runner(request):
    """Uses the default runner"""
    return CliRunner()


def pytest_collection_modifyitems(items: t.Iterable[pytest.Function]):
    """
    Change the default test execution order
    Fundamental tests should be executed first
    """

    # The order of the test function names specifies the order the tests should be executed
    test_function_names = [
        "test_type_assignment",
        "test_widget_registry_command_names",
        "test_type_assignment_multiple_options",
        "test_type_assignment_multiple_commands",
    ]
    # First list should be at the front, the second one after the first one and so on
    move_to_front: list[list[pytest.Function]] = [
        [] for _ in range(len(test_function_names))
    ]

    items_removed = 0
    for i, item in enumerate(
        items[:]
    ):  # Iterate over a copy, because we are modifying the original list
        for j, name in enumerate(test_function_names):
            if (
                re.compile(f"^{name}(\\[|$)").search(item.name) is not None
            ):  # Due to parametrizied tests, the test names contain  (among other things) a "[" after the function name
                move_to_front[j].append(items.pop(i - items_removed))
                items_removed += 1
                break  # inner loop

    for test_functions_list in reversed(
        move_to_front
    ):  # First list should be at the front, so we need to insert the list (of lists) in reverse order
        for test_function in test_functions_list:
            items.insert(0, test_function)



from __future__ import annotations

import typing as t

import click
import pytest

from tests.testutils import ClickAttrs
import clickqt.widgets


@pytest.mark.parametrize(
    ("click_attrs_list", "group_name", "cli_names_list", "expected"),
    [
        (
            [
                [ClickAttrs.checkbox(), ClickAttrs.intfield()],
                [ClickAttrs.realfield(), ClickAttrs.passwordfield()],
            ],
            "test_group",
            ["cli1", "cli2"],
            ["test_group:cli1", "test_group:cli2"],
        ),
        (
            [
                [ClickAttrs.filefield(), ClickAttrs.filepathfield()],
                [ClickAttrs.tuple_widget(types=(click.types.Path(), int))],
            ],
            "abc",
            ["abc", "abc23"],
            ["abc:abc", "abc:abc23"],
        ),
    ],
)
def test_widget_registry_command_names(
    click_attrs_list: t.Iterable[t.Iterable[dict]],
    group_name: str,
    cli_names_list: t.Iterable[str],
    expected: t.Iterable[str],
):
    clis = []

    for i, cli_params in enumerate(click_attrs_list):
        params = []

        for j, click_attrs in enumerate(cli_params):
            params.append(click.Option(param_decls=["--test" + str(j)], **click_attrs))

        clis.append(click.Command(cli_names_list[i], params=params))

    group = click.Group(group_name, commands=clis)

    control = clickqt.qtgui_from_click(group)
    assert len(control.widget_registry) == len(cli_names_list)
    for i, cli_name in enumerate(control.widget_registry.keys()):
        assert cli_name == expected[i]



from __future__ import annotations

from io import BytesIO, TextIOWrapper
import html

from PySide6.QtWidgets import QPlainTextEdit, QMenu
from PySide6.QtGui import QTextCursor, QContextMenuEvent, QAction, QColor
from PySide6.QtCore import Signal


class OutputStream(TextIOWrapper):
    """Sends the content of **stream** as Html-escaped-text to **output**.

    :param output: The object to which the content of **stream** should be sent
    :param stream: The stream-object from which the content should be taken
    :param color: The display color used in **output**
    """

    def __init__(self, output: "TerminalOutput", stream: TextIOWrapper, color: QColor):
        super().__init__(BytesIO(), "utf-8")
        self.output = output
        self.stream = stream
        self.color = color

    def write(self, message: "bytes | str"):
        """Writes **message** to **output** and utf-8 decoded + Html-escaped to **stream**

        :param message: The message which should be written to **output** and **stream**
        """

        if message:
            message = message.decode("utf-8") if isinstance(message, bytes) else message
            print(message, file=self.stream, end="")  # Write to "normal" stream as well
            message = (
                html.escape(message).replace("\r\n", "\n").replace("\n", "<br>")
            )  # Replace '\n' with HTML code

            # Send new message to main thread because worker thread could also be here (-> program crash otherwise)
            self.output.newHtmlMessage.emit(
                f"<p span style='color: rgb({self.color.red()}, {self.color.green()}, {self.color.blue()})'>{message}</p>"
            )


class TerminalOutput(QPlainTextEdit):
    """Displays the output on the screen. Extends the standard context menu with a 'clear'-function."""

    newHtmlMessage: Signal = Signal(
        str
    )  #: Internal Qt-Signal, which will be emitted when there is a new Html-message to display

    def contextMenuEvent(self, event: QContextMenuEvent):  # pragma: no cover
        """Inherited from :class:`~PySide6.QtWidgets.QPlainTextEdit`\n
        Extends the standard context menu with a 'clear'-function.
        """

        menu: QMenu = self.createStandardContextMenu()
        action = QAction("Clear")
        menu.addAction(action)
        action.triggered.connect(self.clear)
        menu.exec(event.globalPos())

    def writeHtml(self, message: str):
        """Appends **message** to the end of the current content.

        :param message: The message in Html-format that should be appended to the end of the current content
        """

        self.moveCursor(QTextCursor.End)
        self.textCursor().insertHtml(message)



from __future__ import annotations

import os
import typing as t

import click
import pytest

from tests.testutils import ClickAttrs
import clickqt.widgets


@pytest.mark.parametrize(
    ("click_attrs", "envvar_values", "expected"),
    [
        (ClickAttrs.textfield(), "test123", "test123"),
        (
            ClickAttrs.textfield(),
            ["test1", "test2"],
            os.path.pathsep.join(["test1", "test2"]),
        ),
        (ClickAttrs.filefield(), "test", "test"),
        (
            ClickAttrs.filefield(),
            ["test1", "test2"],
            os.path.pathsep.join(["test1", "test2"]),
        ),
        (ClickAttrs.filepathfield(), "test", "test"),
        (
            ClickAttrs.filepathfield(),
            ["test1", "test2"],
            os.path.pathsep.join(["test1", "test2"]),
        ),
        (
            ClickAttrs.multi_value_widget(nargs=2, type=click.types.Path()),
            ["a", "b"],
            ["a", "b"],
        ),
        (
            ClickAttrs.multi_value_widget(nargs=2, type=click.types.File()),
            ["a", "b"],
            ["a", "b"],
        ),
        (
            ClickAttrs.intfield(),
            "test123",
            0,
        ),  # envvars are only considered for string based widgets
        (
            ClickAttrs.multi_value_widget(nargs=2, type=int),
            ["2", "3"],
            [2, 3],
        ),  # and for multi widgets
        (ClickAttrs.tuple_widget(types=(float, str)), ["2.3", "3"], [2.3, "3"]),
        (ClickAttrs.nvalue_widget(), ["2.3", "3", "av"], ["2.3", "3", "av"]),
    ],
)
def test_set_envvar(
    click_attrs: dict,
    envvar_values: "str| t.Sequence[str]",
    expected: "str| t.Sequence[t.Any]",
):
    os.environ["TEST_CLICKQT_ENVVAR"] = os.path.pathsep.join(
        envvar_values
        if isinstance(envvar_values, t.Sequence) and not isinstance(envvar_values, str)
        else [envvar_values]
    )

    param = click.Option(
        param_decls=["--test"], envvar="TEST_CLICKQT_ENVVAR", **click_attrs
    )
    cli = click.Command("cli", params=[param])

    control = clickqt.qtgui_from_click(cli)
    assert control.widget_registry[cli.name][param.name].get_widget_value() == expected


@pytest.mark.parametrize(
    ("click_attrs", "envvar_values", "expected"),
    [
        (
            ClickAttrs.multi_value_widget(nargs=3, type=click.types.Path()),
            ["a", "b"],
            "Takes 3 values but 2 were given.",
        ),
        (
            ClickAttrs.multi_value_widget(nargs=2, type=click.types.File()),
            ["a", "b", "c"],
            "Takes 2 values but 3 were given.",
        ),
        (
            ClickAttrs.tuple_widget(types=(click.types.Path(), click.types.Path())),
            ["a", "b", "c"],
            "Takes 2 values but 3 were given.",
        ),
        (
            ClickAttrs.tuple_widget(types=(click.types.Path(), click.types.File())),
            ["a"],
            "Takes 2 values but 1 was given.",
        ),
        (
            ClickAttrs.tuple_widget(types=(click.types.File(), click.types.Path())),
            ["a", "c", "b"],
            "Takes 2 values but 3 were given.",
        ),
        (
            ClickAttrs.tuple_widget(types=(click.types.File(), click.types.File())),
            ["a", "b", "c", "d"],
            "Takes 2 values but 4 were given.",
        ),
    ],
)
def test_set_envvar_fail(
    click_attrs: dict,
    envvar_values: t.Union[str, t.Sequence[str]],
    expected: t.Union[str, t.Sequence[str]],
):
    os.environ["TEST_CLICKQT_ENVVAR"] = os.path.pathsep.join(
        envvar_values if isinstance(envvar_values, t.Sequence) else [envvar_values]
    )

    param = click.Option(
        param_decls=["--test"], envvar="TEST_CLICKQT_ENVVAR", **click_attrs
    )
    cli = click.Command("cli", params=[param])

    with pytest.raises(click.exceptions.BadParameter) as exc_info:
        clickqt.qtgui_from_click(cli).widget_registry[cli.name][param.name]

    assert expected == exc_info.value.message



import os
import click
import clickqt


@click.group()
def utilgroup():
    pass


@utilgroup.command()
@click.argument("username", default=lambda: os.environ.get("USERNAME", ""))
@click.option(
    "--verbose",
    type=bool,
    is_flag=True,
    default=True,
    required=True,
    help="Verbosity of the output",
)
@click.option("-c", "--count", count=True, default=3, help="Repetition of the option")
@click.option(
    "--hash-type-single", confirmation_prompt=True, type=click.Choice(["MD5", "SHA1"])
)
@click.option(
    "--hash-type-multiple",
    required=True,
    type=click.Choice(["MD5", "SHA1"]),
    multiple=True,
)
@click.option(
    "-r",
    "--range",
    confirmation_prompt=True,
    type=click.FloatRange(max=20.23, clamp=True),
)
@click.password_option()
@click.confirmation_option(
    expose_value=False,
    prompt="Are you sure you want to run the application with these options?",
)
@click.argument("filename", type=click.Path(exists=True))
@click.argument("input", type=click.File("rb"))
@click.argument("output", type=click.File("wb"))
def test(
    verbose,
    username,
    count,
    hash_type_single,
    hash_type_multiple,
    range,
    password,
    filename,
    input,
    output,
):
    click.echo(
        f"verbose: '{verbose}'\n"
        + f"username: '{username}'\n"
        + f"count: '{count}'\n"
        + f"hash_type_single: '{hash_type_single}'\n"
        + f"hash_type_multiple: '{hash_type_multiple}'\n"
        + f"range: '{range}'\n"
        + f"password: '{password}'\n"
        + f"filename: '{filename}'"
    )
    click.echo("input: ", nl=False)
    while True:
        chunk = input.read(1024)
        if not chunk:
            break
        output.write(chunk)
    click.echo()  # New line


@utilgroup.command()
@click.option(
    "--userinfo",
    type=(str, int, click.types.DateTime()),
    default=["test", 1, "2023-06-14 15:20:25"],
)
def greet(userinfo):
    fname, no, date = userinfo
    date = date.strftime("%Y-%m-%d")
    click.echo(f"Hello, {fname}! Int, Date: {no, date}.")


@utilgroup.command()
@click.option("--pos", type=int, nargs=2, default=[1, 2])
def position(pos):
    a, b = pos
    click.echo(f"{a}/{b}")


@click.group()
def hello():
    print("Hello group")


@hello.command()
@click.option("--n", type=int, default=3)
def test(n):
    for i in range(n):
        click.echo(i)


def test_callback(ctx, param, value):
    if value == "2":
        return "123"
    else:
        raise click.BadParameter(f"Wrong value")


@hello.command()
@click.option("-t", type=str, callback=test_callback)
def test_with_callback(t):
    click.echo(t)


@hello.command()
@click.option(
    "-ns", type=(int, str), multiple=True, required=True, default=[(1, "a"), (2, "b")]
)
def hello_ns(ns):
    for i, s in ns:
        for _ in range(i):
            click.echo(f"{s}{i}")


@hello.command()
@click.option(
    "paths",
    "--path",
    multiple=True,
    envvar="PATH",
    default=[".", "tests"],
    type=click.Path(exists=True),
)
def hellp_path(paths):
    for path in paths:
        click.echo(path)


@click.group()
def hello2():
    print("Hello2 group")


@hello2.command()
@click.option("-ns", type=(int, str), multiple=True)
def hello_ns2(ns):
    for i, s in ns:
        for _ in range(i):
            click.echo(f"{s}{i}")


utilgroup.add_command(hello)
hello.add_command(hello2)

gui = clickqt.qtgui_from_click(utilgroup)

if __name__ == "__main__":
    utilgroup()



from __future__ import annotations

import sys

from PySide6.QtWidgets import QSpinBox, QDoubleSpinBox
import click

from clickqt.widgets.basewidget import NumericField, BaseWidget


class IntField(NumericField):
    """Represents a click.types.IntParamType- and click.types.IntRange-object.
    If the click object is of type `click.types.IntParamType`, the minimum will be set to -2**31 and the maximum to 2**31 - 1.
    In the other case, the minimum and maximum are defined by clicks range options.

    :param otype: The type which specifies the clickqt widget type. This type may be different compared to **param**.type when dealing with click.types.CompositeParamType-objects
    :param param: The parameter from which **otype** came from
    :param kwargs: Additionally parameters ('widgetsource', 'parent', 'com', 'label') needed for
                    :class:`~clickqt.widgets.basewidget.MultiWidget`- / :class:`~clickqt.widgets.confirmationwidget.ConfirmationWidget`-widgets
    """

    widget_type = QSpinBox  #: The Qt-type of this widget.

    def __init__(self, otype: click.ParamType, param: click.Parameter, **kwargs):
        super().__init__(otype, param, **kwargs)

        assert isinstance(
            otype, (click.IntRange, type(click.INT))
        ), f"'otype' must be of type '{click.IntRange}' or '{type(click.INT)}', but is '{type(otype)}'."

        if not isinstance(otype, click.IntRange):
            # QSpinBox is limited to [-2**31; 2**31 - 1], but sys.maxsize returns 2**63 - 1
            self.set_minimum(-(2**31))  # Default is 0
            self.set_maximum(2**31 - 1)  # Default is 99
        else:
            if otype.min is not None:
                self.set_minimum(otype.min + (1 if otype.min_open else 0))
            if otype.max is not None:
                self.set_maximum(otype.max - (1 if otype.max_open else 0))

        if self.parent_widget is None:
            self.set_value(BaseWidget.get_param_default(param, 0))


class RealField(NumericField):
    """Represents a click.types.FloatParamType- and click.types.FloatRange-object. Two decimals will be displayed by default.
    If the click object is of type `click.types.FloatParamType`, the minimum will be set to -sys.float_info.max and the maximum to sys.float_info.max.
    In the other case, the minimum and maximum are defined by clicks range options.

    :param otype: The type which specifies the clickqt widget type. This type may be different compared to **param**.type when dealing with click.types.CompositeParamType-objects
    :param param: The parameter from which **otype** came from
    :param kwargs: Additionally parameters ('widgetsource', 'parent', 'com', 'label') needed for
                    :class:`~clickqt.widgets.basewidget.MultiWidget`- / :class:`~clickqt.widgets.confirmationwidget.ConfirmationWidget`-widgets
    """

    widget_type = QDoubleSpinBox  #: The Qt-type of this widget.

    def __init__(self, otype: click.ParamType, param: click.Parameter, **kwargs):
        super().__init__(otype, param, **kwargs)

        assert isinstance(
            otype, (click.FloatRange, type(click.FLOAT))
        ), f"'otype' must be of type '{click.FloatRange}' or '{type(click.FLOAT)}', but is '{type(otype)}'."

        if not isinstance(otype, click.FloatRange):
            self.set_minimum(-sys.float_info.max)  # Default is 0.0
            self.set_maximum(sys.float_info.max)  # Default is 99.0
        else:
            if otype.min is not None:
                self.set_minimum(
                    otype.min
                    + (10 ** -self.widget.decimals() if otype.min_open else 0.0)
                )
            if otype.max is not None:
                self.set_maximum(
                    otype.max
                    - (10 ** -self.widget.decimals() if otype.max_open else 0.0)
                )

        if self.parent_widget is None:
            self.set_value(BaseWidget.get_param_default(param, 0.0))



""" Contains the FilePathField class """
from __future__ import annotations

import click
from PySide6.QtWidgets import QLineEdit

from clickqt.widgets.textfield import PathField


class FilePathField(PathField):
    """Represents a click.types.Path object.

    :param otype: The type which specifies the clickqt widget type.
        This type may be different compared to **param**.type when dealing with click.types.CompositeParamType-objects
    :param param: The parameter from which **otype** came from
    :param kwargs: Additionally parameters ('parent', 'widgetsource', 'com', 'label') needed for
        :class:`~clickqt.widgets.basewidget.MultiWidget`- /
        :class:`~clickqt.widgets.confirmationwidget.ConfirmationWidget`- widgets
    """

    widget_type = QLineEdit  #: The Qt-type of this widget.

    def __init__(self, otype: click.ParamType, param: click.Parameter, **kwargs):
        super().__init__(otype, param, **kwargs)

        assert isinstance(
            otype, click.Path
        ), f"'otype' must be of type '{click.Path}', but is '{type(otype)}'."

        #: File type is
        #   :attr:`~clickqt.widgets.textfield.PathField.FileType.File`
        # and/or
        #   :attr:`~clickqt.widgets.textfield.PathField.FileType.Directory`
        #: depending on **otype**\.file_okay and **otype**\.dir_okay.
        self.file_type: PathField.FileType
        self.file_type |= PathField.FileType.File if otype.file_okay else self.file_type
        self.file_type |= (
            PathField.FileType.Directory if otype.dir_okay else self.file_type
        )

        assert (
            self.file_type != PathField.FileType.Unknown
        ), f"Neither 'file_okay' nor 'dir_okay' in option '{self.widget_name}' is set"



from __future__ import annotations

from abc import ABC, abstractmethod
import os
import typing as t
from gettext import ngettext

from PySide6.QtWidgets import QWidget, QLabel, QVBoxLayout, QHBoxLayout
from PySide6.QtCore import Qt
import click

from clickqt.core.error import ClickQtError
import clickqt.core  # FocusOutValidator


class BaseWidget(ABC):
    """Provides basic functionalities and initializes the widget.
    Every clickqt widget has to inherit from this class.

    :param otype: The type which specifies the clickqt widget type. This type may be different compared to **param**.type when dealing with click.types.CompositeParamType-objects
    :param param: The parameter from which **otype** came from
    :param parent: The parent BaseWidget of **otype**, defaults to None. Needed for :class:`~clickqt.widgets.basewidget.MultiWidget`-widgets
    :param kwargs: Additionally parameters ('widgetsource', 'com', 'label') needed for
                    :class:`~clickqt.widgets.basewidget.MultiWidget`- / :class:`~clickqt.widgets.confirmationwidget.ConfirmationWidget`-widgets
    """

    widget_type: t.ClassVar[t.Type]  #: The Qt-type of this widget.

    def __init__(
        self,
        otype: click.ParamType,
        param: click.Parameter,
        parent: t.Optional["BaseWidget"] = None,
        **kwargs,
    ):
        assert isinstance(otype, click.ParamType)
        assert isinstance(param, click.Parameter)
        self.type = otype
        self.param = param
        self.parent_widget = parent
        self.click_command: click.Command = kwargs.get("com")
        self.widget_name = param.name
        self.container = QWidget()
        self.layout = (
            QVBoxLayout()
            if parent is None or kwargs.get("vboxlayout")
            else QHBoxLayout()
        )

        self.label = QLabel(text=f"<b>{kwargs.get('label', '')}{self.widget_name}</b>")
        self.label.setTextFormat(Qt.TextFormat.RichText)  # Bold text

        self.widget = self.create_widget()

        self.layout.addWidget(self.label)
        if (
            isinstance(param, click.Option)
            and param.help
            and (parent is None or kwargs.get("vboxlayout"))
        ):  # Help text
            help_label = QLabel(text=param.help)
            help_label.setWordWrap(True)  # Multi-line
            self.layout.addWidget(help_label)
        self.layout.addWidget(self.widget)
        self.container.setLayout(self.layout)

        self.widget.setObjectName(
            param.name
        )  # Only change the stylesheet of this widget and not of all (child-)widgets

        assert self.widget is not None, "Widget not initialized"
        assert self.param is not None, "Click param object not provided"
        assert self.click_command is not None, "Click command not provided"
        assert self.type is not None, "Type not provided"

        self.focus_out_validator = clickqt.core.FocusOutValidator(self)
        self.widget.installEventFilter(self.focus_out_validator)

    def create_widget(self) -> QWidget:
        """Creates the widget specified in :attr:`~clickqt.widgets.basewidget.BaseWidget.widget_type` and returns it."""

        return self.widget_type()

    @abstractmethod
    def set_value(self, value: t.Any):
        """Sets the value of the Qt-widget.

        :param value: The new value that should be stored in the widget
        :raises click.BadParameter: **value** could not be converted into the corresponding click.ParamType
        """

    def is_empty(self) -> bool:
        """Checks whether the widget is empty. This can be the case for string-based widgets or the
        multiple choice-widget (:class:`~clickqt.widgets.combobox.CheckableComboBox`).\n
        Subclasses may need to override this method.

        :return: False
        """
        return False

    def get_value(self) -> tuple[t.Any, ClickQtError]:
        """Validates the value of the Qt-widget and returns the result.

        :return: Valid: (widget value or the value of a callback, :class:`~clickqt.core.error.ClickQtError.ErrorType.NO_ERROR`)\n
                 Invalid: (None, :class:`~clickqt.core.error.ClickQtError.ErrorType.CONVERTING_ERROR` or
                 :class:`~clickqt.core.error.ClickQtError.ErrorType.PROCESSING_VALUE_ERROR` or :class:`~clickqt.core.error.ClickQtError.ErrorType.REQUIRED_ERROR`)
        """
        value: t.Any = None

        # Try to convert the provided value into the corresponding click object type
        try:  # pylint: disable=too-many-try-statements, too-many-nested-blocks
            default = BaseWidget.get_param_default(self.param, None)
            # if statement is obtained by creating the corresponding truth table
            if self.param.multiple or (
                not isinstance(self.type, click.Tuple) and self.param.nargs != 1
            ):
                value_missing = False
                widget_values: list = self.get_widget_value()

                if len(widget_values) == 0:  # Checkable combobox
                    if self.param.required and default is None:
                        self.handle_valid(False)
                        return (
                            None,
                            ClickQtError(
                                ClickQtError.ErrorType.REQUIRED_ERROR,
                                self.widget_name,
                                self.param.param_type_name,
                            ),
                        )
                    #self.handle_parameter_missing_default(default)
                    if default is not None:
                        self.set_value(default)
                        widget_values = self.get_widget_value()
                    else:  # param is not required and there is no default -> value is None
                        value_missing = True  # But callback should be considered

                if not value_missing:
                    value = []
                    for i, v in enumerate(
                        widget_values
                    ):  # v is not a BaseWidget, but a primitive type
                        if (
                            str(v) == ""
                        ):  # Empty widget (only possible for string based widgets)
                            if self.param.required and default is None:
                                self.handle_valid(False)
                                return (
                                    None,
                                    ClickQtError(
                                        ClickQtError.ErrorType.REQUIRED_ERROR,
                                        self.widget_name,
                                        self.param.param_type_name,
                                    ),
                                )
                            if default is not None and i < len(
                                default
                            ):  # Overwrite the empty widget with the default value and execute with this (new) value
                                values = self.get_widget_value()
                                values[i] = default[
                                    i
                                ]  # Only overwrite the empty widget, not all
                                self.set_value(values)
                                v = default[i]
                            else:  # param is not required, widget is empty and there is no default (click equivalent: option not provided in click command cmd)
                                value = None
                                break

                        value.append(
                            self.type.convert(
                                value=v,
                                param=self.param,
                                ctx=click.Context(self.click_command),
                            )
                        )
            else:
                value_missing = False
                if self.is_empty():
                    if self.param.required and default is None:
                        self.handle_valid(False)
                        return (
                            None,
                            ClickQtError(
                                ClickQtError.ErrorType.REQUIRED_ERROR,
                                self.widget_name,
                                self.param.param_type_name,
                            ),
                        )
                    if default is not None:
                        self.set_value(default)
                    else:
                        value_missing = True  # -> value is None

                if not value_missing:
                    value = self.type.convert(
                        value=self.get_widget_value(),
                        param=self.param,
                        ctx=click.Context(self.click_command),
                    )
        except Exception as e: # pylint: disable=broad-exception-caught
            self.handle_valid(False)
            return (
                None,
                ClickQtError(
                    ClickQtError.ErrorType.CONVERTING_ERROR, self.widget_name, e
                ),
            )

        return self.handle_callback(value)

    def handle_callback(self, value: t.Any) -> tuple[t.Any, ClickQtError]:
        """Validates **value** in the user-defined callback (if provided) and returns the result.

        :param value: The value that should be validated in the callback

        :return: Valid: (**value** or the value of a callback, :class:`~clickqt.core.error.ClickQtError.ErrorType.NO_ERROR`)\n
                 Invalid: (None, :class:`~clickqt.core.error.ClickQtError.ErrorType.ABORTED_ERROR` or
                 :class:`~clickqt.core.error.ClickQtError.ErrorType.EXIT_ERROR` or :class:`~clickqt.core.error.ClickQtError.ErrorType.PROCESSING_VALUE_ERROR`)
        """

        try:  # Consider callbacks
            ret_val = (
                self.param.process_value(click.Context(self.click_command), value),
                ClickQtError(),
            )
            self.handle_valid(True)
            return ret_val
        except click.exceptions.Abort:
            return (None, ClickQtError(ClickQtError.ErrorType.ABORTED_ERROR))
        except click.exceptions.Exit:
            return (None, ClickQtError(ClickQtError.ErrorType.EXIT_ERROR))
        except Exception as e: # pylint: disable=broad-exception-caught
            self.handle_valid(False)
            return (
                None,
                ClickQtError(
                    ClickQtError.ErrorType.PROCESSING_VALUE_ERROR, self.widget_name, e
                ),
            )

    @abstractmethod
    def get_widget_value(self) -> t.Any:
        """Returns the value of the Qt-widget without any checks."""

    def handle_valid(self, valid: bool):
        """Changes the border of the widget dependent on **valid**. If **valid** == False, the border will be colored red, otherwise black.

        :param valid: Specifies whether there was no error when validating the widget

        """

        if not valid:
            self.widget.setStyleSheet(
                f"QWidget#{self.widget.objectName()}{{ border: 1px solid red }}"
            )
        else:
            self.widget.setStyleSheet(f"QWidget#{self.widget.objectName()}{{ }}")

    @staticmethod
    def get_param_default(param: click.Parameter, alternative:t.Any=None):
        """Returns the default value of **param**. If there is no default value, **alternative** will be returned."""

        # TODO: Replace with param.get_default(ctx=click.Context(command), call=True)
        if param.default is None:
            return alternative
        if callable(param.default):
            return param.default()
        return param.default


class NumericField(BaseWidget):
    """Provides basic functionalities for numeric based widgets

    :param otype: The type which specifies the clickqt widget type. This type may be different compared to **param**.type when dealing with click.types.CompositeParamType-objects
    :param param: The parameter from which **otype** came from
    :param kwargs: Additionally parameters ('parent', 'widgetsource', 'com', 'label') needed for
                    :class:`~clickqt.widgets.basewidget.MultiWidget`- / :class:`~clickqt.widgets.confirmationwidget.ConfirmationWidget`-widgets
    """

    def set_value(self, value: t.Any):
        self.widget.setValue(
            self.type.convert(
                value=str(value),
                param=self.click_command,
                ctx=click.Context(self.click_command),
            )
        )

    def set_minimum(self, minval: t.Union[int, float]):
        """Sets the minimum value."""

        self.widget.setMinimum(minval)

    def set_maximum(self, maxval: t.Union[int, float]):
        """Sets the maximum value."""

        self.widget.setMaximum(maxval)

    def get_widget_value(self) -> t.Union[int, float]:
        return self.widget.value()


class ComboBoxBase(BaseWidget):
    """Provides basic functionalities for click.types.Choice based widgets

    :param otype: The type which specifies the clickqt widget type. This type may be different compared to **param**.type when dealing with click.types.CompositeParamType-objects
    :param param: The parameter from which **otype** came from
    :param kwargs: Additionally parameters ('parent', 'widgetsource', 'com', 'label') needed for
                    :class:`~clickqt.widgets.basewidget.MultiWidget`- / :class:`~clickqt.widgets.confirmationwidget.ConfirmationWidget`-widgets
    """

    def __init__(self, otype: click.ParamType, param: click.Parameter, **kwargs):
        super().__init__(otype, param, **kwargs)

        assert isinstance(
            otype, click.Choice
        ), f"'otype' must be of type '{click.Choice}', but is '{type(otype)}'."

        self.add_items(otype.choices)

    @abstractmethod
    def add_items(self, items: t.Iterable[str]):
        """Adds each of the strings in **items** to the checkable combobox."""


class MultiWidget(BaseWidget):
    """Provides basic functionalities for click.types.CompositeParamType based widgets and multi value widgets.

    :param otype: The type which specifies the clickqt widget type. This type may be different compared to **param**.type when dealing with click.types.CompositeParamType-objects
    :param param: The parameter from which **otype** came from
    :param kwargs: Additionally parameters ('parent', 'widgetsource', 'com', 'label') needed for
                    :class:`~clickqt.widgets.basewidget.MultiWidget`- / :class:`~clickqt.widgets.confirmationwidget.ConfirmationWidget`-widgets
    """

    def __init__(self, otype: click.ParamType, param: click.Parameter, **kwargs):
        super().__init__(otype, param, **kwargs)

        self.children: t.Union[t.Iterable[BaseWidget],t.dict_values[BaseWidget]] = []

    def init(self):
        """Sets the value of the (child-)widgets according to envvar or default values.
        If the envvar values are None, the defaults values will be considered.
        """

        if self.parent_widget is None:
            # Consider envvar
            if (
                envvar_values := self.param.resolve_envvar_value(
                    click.Context(self.click_command)
                )
            ) is not None:
                # self.type.split_envvar_value(envvar_values) does not work because clicks "self.envvar_list_splitter" is not set corrently
                self.set_value(envvar_values.split(os.path.pathsep))
            elif (
                default := BaseWidget.get_param_default(self.param, None)
            ) is not None:  # Consider default value
                self.set_value(default)

    def set_value(self, value: t.Iterable[t.Any]):
        if len(value) != self.param.nargs:
            raise click.BadParameter(
                ngettext(
                    "Takes {nargs} values but 1 was given.",
                    "Takes {nargs} values but {len} were given.",
                    len(value),
                ).format(nargs=self.param.nargs, len=len(value)),
                ctx=click.Context(self.click_command),
                param=self.param,
            )

        for i, c in enumerate(self.children):
            c.set_value(value[i])

    def handle_valid(self, valid: bool):
        for c in self.children:
            c.handle_valid(valid)

    def is_empty(self) -> bool:
        """ "Checks whether the widget is empty. This is the case when there are no children or when at least one (string-based) children is empty.

        :return: True, if this widget has no children or at least one children is empty, False otherwise
        """

        if len(self.children) == 0:
            return True

        return any([c.is_empty() for c in self.children])

    def get_widget_value(self) -> t.Iterable[t.Any]:
        return [c.get_widget_value() for c in self.children]



from __future__ import annotations

import sys
import typing as t

import click
import pytest
from PySide6.QtWidgets import (
    QLineEdit,
    QFileDialog,
    QApplication,
    QPushButton,
    QMessageBox,
)
from PySide6.QtCore import QTimer, Signal, QObject, Qt
from pytestqt.qtbot import QtBot

from tests.testutils import ClickAttrs
import clickqt.widgets


class CustomParamType(click.ParamType):
    pass


@pytest.mark.parametrize(
    ("click_attrs", "expected_clickqt_type"),
    [
        (ClickAttrs.checkbox(), clickqt.widgets.CheckBox),
        (ClickAttrs.messagebox(prompt="Test"), clickqt.widgets.MessageBox),
        (ClickAttrs.intfield(), clickqt.widgets.IntField),
        (ClickAttrs.realfield(), clickqt.widgets.RealField),
        (ClickAttrs.confirmation_widget(), clickqt.widgets.ConfirmationWidget),
        (ClickAttrs.textfield(), clickqt.widgets.TextField),
        (ClickAttrs.passwordfield(), clickqt.widgets.PasswordField),
        (ClickAttrs.datetime(), clickqt.widgets.DateTimeEdit),
        (ClickAttrs.uuid(), clickqt.widgets.TextField),
        (ClickAttrs.unprocessed(), clickqt.widgets.TextField),
        (ClickAttrs.combobox(choices=["a"]), clickqt.widgets.ComboBox),
        (
            ClickAttrs.checkable_combobox(choices=["a"]),
            clickqt.widgets.CheckableComboBox,
        ),
        (ClickAttrs.intrange(), clickqt.widgets.IntField),
        (ClickAttrs.floatrange(), clickqt.widgets.RealField),
        (ClickAttrs.filefield(), clickqt.widgets.FileField),
        (ClickAttrs.filepathfield(), clickqt.widgets.FilePathField),
        (ClickAttrs.nvalue_widget(), clickqt.widgets.NValueWidget),
        (
            ClickAttrs.tuple_widget(types=(click.types.Path(), int)),
            clickqt.widgets.TupleWidget,
        ),
        (ClickAttrs.multi_value_widget(nargs=2), clickqt.widgets.MultiValueWidget),
        (
            {"type": CustomParamType()},
            clickqt.widgets.TextField,
        ),  # Custom types are mapped to TextFields
    ],
)
def test_type_assignment(
    click_attrs: dict, expected_clickqt_type: clickqt.widgets.BaseWidget
):
    param = click.Option(param_decls=["--test"], **click_attrs)
    cli = click.Command("cli", params=[param])

    control = clickqt.qtgui_from_click(cli)
    gui = control.gui

    assert isinstance(
        gui.create_widget(param.type, param, widgetsource=gui.create_widget, com=cli),
        expected_clickqt_type,
    ), "directly"  # Perfect type match
    assert isinstance(
        control.widget_registry[cli.name][param.name], expected_clickqt_type
    ), "clickqt"  # Perfect type match


@pytest.mark.parametrize(
    ("value"),
    [
        ("upper"),
        ("lower"),
    ],
)
def test_feature_switch(value: str):
    param1 = click.Option(param_decls=["--upper", "transformation"], flag_value="upper")
    param2 = click.Option(param_decls=["--lower", "transformation"], flag_value="lower")
    cli = click.Command("cli", params=[param1, param2])

    control = clickqt.qtgui_from_click(cli)
    widget = control.widget_registry[cli.name][param1.name]

    assert isinstance(
        widget, clickqt.widgets.ComboBox
    )  # Feature switches are mapped to ComboBoxes

    widget.set_value(value)

    assert widget.get_widget_value() == value


@pytest.mark.parametrize(
    ("click_attrs_list", "expected_clickqt_type_list"),
    [
        (
            [
                ClickAttrs.checkbox(),
                ClickAttrs.intfield(),
                ClickAttrs.realfield(),
                ClickAttrs.passwordfield(),
            ],
            [
                clickqt.widgets.CheckBox,
                clickqt.widgets.IntField,
                clickqt.widgets.RealField,
                clickqt.widgets.PasswordField,
            ],
        ),
        (
            [
                ClickAttrs.filefield(),
                ClickAttrs.filepathfield(),
                ClickAttrs.tuple_widget(types=(click.types.Path(), int)),
            ],
            [
                clickqt.widgets.FileField,
                clickqt.widgets.FilePathField,
                clickqt.widgets.TupleWidget,
            ],
        ),
        (
            [
                ClickAttrs.datetime(),
                ClickAttrs.combobox(choices=["a"]),
                ClickAttrs.nvalue_widget(),
                ClickAttrs.multi_value_widget(nargs=2),
            ],
            [
                clickqt.widgets.DateTimeEdit,
                clickqt.widgets.ComboBox,
                clickqt.widgets.NValueWidget,
                clickqt.widgets.MultiValueWidget,
            ],
        ),
    ],
)
def test_type_assignment_multiple_options(
    click_attrs_list: t.Iterable[dict],
    expected_clickqt_type_list: t.Iterable[clickqt.widgets.BaseWidget],
):
    params = []

    for i, click_attrs in enumerate(click_attrs_list):
        params.append(click.Option(param_decls=["--test" + str(i)], **click_attrs))

    cli = click.Command("cli", params=params)

    control = clickqt.qtgui_from_click(cli)
    for i, v in enumerate(control.widget_registry[cli.name].values()):
        # Perfect type match
        # pylint: disable=unidiomatic-typecheck
        assert type(v) is expected_clickqt_type_list[i]


@pytest.mark.parametrize(
    ("click_attrs_list", "expected_clickqt_type_list"),
    [
        (
            [
                [ClickAttrs.checkbox(), ClickAttrs.intfield()],
                [ClickAttrs.realfield(), ClickAttrs.passwordfield()],
            ],
            [
                [clickqt.widgets.CheckBox, clickqt.widgets.IntField],
                [clickqt.widgets.RealField, clickqt.widgets.PasswordField],
            ],
        ),
        (
            [
                [ClickAttrs.filefield(), ClickAttrs.filepathfield()],
                [ClickAttrs.tuple_widget(types=(click.types.Path(), int))],
            ],
            [
                [clickqt.widgets.FileField, clickqt.widgets.FilePathField],
                [clickqt.widgets.TupleWidget],
            ],
        ),
        (
            [
                [
                    ClickAttrs.datetime(),
                    ClickAttrs.combobox(choices=["a"]),
                    ClickAttrs.nvalue_widget(),
                ],
                [ClickAttrs.multi_value_widget(nargs=2)],
            ],
            [
                [
                    clickqt.widgets.DateTimeEdit,
                    clickqt.widgets.ComboBox,
                    clickqt.widgets.NValueWidget,
                ],
                [clickqt.widgets.MultiValueWidget],
            ],
        ),
    ],
)
def test_type_assignment_multiple_commands(
    click_attrs_list: t.Iterable[t.Iterable[dict]],
    expected_clickqt_type_list: t.Iterable[t.Iterable[clickqt.widgets.BaseWidget]],
):
    clis = []

    for i, cli_params in enumerate(click_attrs_list):
        params = []

        for j, click_attrs in enumerate(cli_params):
            params.append(click.Option(param_decls=["--test" + str(j)], **click_attrs))

        clis.append(click.Command("cli" + str(i), params=params))

    group = click.Group("group", commands=clis)

    control = clickqt.qtgui_from_click(group)
    for i, cli_name in enumerate(control.widget_registry.keys()):
        for j, v in enumerate(control.widget_registry[cli_name].values()):
            # pylint disable=unidiomatic-typecheck
            assert type(v) is expected_clickqt_type_list[i][j]  # Perfect type match


def test_passwordfield_showPassword():
    param = click.Option(param_decls=["--p"], **ClickAttrs.passwordfield())
    cli = click.Command("cli", params=[param])

    control = clickqt.qtgui_from_click(cli)
    passwordfield_widget: clickqt.widgets.PasswordField = control.widget_registry[
        cli.name
    ][param.name]

    for _ in range(3):
        assert (
            passwordfield_widget.widget.echoMode() == QLineEdit.EchoMode.Normal
            if passwordfield_widget.show_hide_action.isChecked()
            else QLineEdit.EchoMode.Password
        )
        # QIcons cannot be compared, but QImages can
        icon = passwordfield_widget.show_hide_action.icon()
        expected_icon = passwordfield_widget.icon_text[
            passwordfield_widget.show_hide_action.isChecked()
        ][0]
        assert (
            icon.pixmap(176, 176).toImage() == expected_icon.pixmap(176, 176).toImage()
        )
        assert (
            passwordfield_widget.show_hide_action.text()
            == passwordfield_widget.icon_text[
                passwordfield_widget.show_hide_action.isChecked()
            ][1]
        )

        passwordfield_widget.show_hide_action.setChecked(
            not passwordfield_widget.show_hide_action.isChecked()
        )


@pytest.mark.skipif(
    sys.platform == "darwin", reason="Not runnable on GitHubs MacOS-VMs (stuck)"
)
@pytest.mark.parametrize(
    ("click_attrs", "value", "expected"),
    [
        pytest.param(
            ClickAttrs.filefield(type_dict={"mode": "r"}),
            ".gitignore",
            ".gitignore",
            marks=pytest.mark.skipif(
                sys.platform == "linux", reason="Does not work under linux"
            ),
        ),
        (ClickAttrs.filefield(type_dict={"mode": "r"}), "invalid_file.txt", ""),
        (ClickAttrs.filefield(type_dict={"mode": "w"}), ".gitignore", ".gitignore"),
        (
            ClickAttrs.filefield(type_dict={"mode": "w"}),
            "invalid_file.txt",
            "invalid_file.txt",
        ),
        (ClickAttrs.filepathfield(type_dict={"exists": True}), "invalid_path", ""),
        (
            ClickAttrs.filepathfield(type_dict={"exists": True}),
            "tests",
            "tests",
        ),  # valid folder
        pytest.param(
            ClickAttrs.filepathfield(type_dict={"exists": True}),
            ".gitignore",
            ".gitignore",
            marks=pytest.mark.skipif(
                sys.platform == "linux", reason="Does not work under linux"
            ),
        ),  # valid file
        (
            ClickAttrs.filepathfield(type_dict={"exists": False}),
            "invalid_path",
            "invalid_path",
        ),  # Exists==False: Accept any file
        pytest.param(
            ClickAttrs.filepathfield(type_dict={"exists": True, "dir_okay": False}),
            ".gitignore",
            ".gitignore",
            marks=pytest.mark.skipif(
                sys.platform == "linux", reason="Does not work under linux"
            ),
        ),
        (
            ClickAttrs.filepathfield(type_dict={"exists": True, "dir_okay": False}),
            "tests",
            "",
        ),
        (
            ClickAttrs.filepathfield(type_dict={"exists": False, "dir_okay": False}),
            "tests",
            "",
        ),
        (
            ClickAttrs.filepathfield(type_dict={"exists": True, "file_okay": False}),
            ".gitignore",
            "",
        ),
        (
            ClickAttrs.filepathfield(type_dict={"exists": True, "file_okay": False}),
            "tests",
            "tests",
        ),
        (
            ClickAttrs.filepathfield(type_dict={"exists": False, "file_okay": False}),
            ".gitignore",
            "",
        ),  # Not a folder
    ],
)
def test_pathfield(qtbot: QtBot, click_attrs: dict, value: str, expected: str):
    param = click.Option(param_decls=["--p"], **click_attrs)
    cli = click.Command("cli", params=[param])

    control = clickqt.qtgui_from_click(cli)
    widget: clickqt.widgets.PathField = control.widget_registry[cli.name][param.name]

    class Finished(QObject):
        finished = Signal()

    def closeMessagebox(message_box_closed: Finished):
        messagebox: QMessageBox = QApplication.activeModalWidget()

        # Wait, until we have the QMessageBox- or QFileDialog-object
        tries = 0
        while (
            messagebox is not None
            and not isinstance(messagebox, (QFileDialog, QMessageBox))
            and tries < 10
        ):
            QApplication.processEvents()
            messagebox = QApplication.activeModalWidget()
            tries += 1

        if messagebox is not None and isinstance(messagebox, QMessageBox):
            messagebox.close()

        message_box_closed.finished.emit()

    def selectFile():
        file_dialog: QFileDialog = QApplication.activeModalWidget()
        message_box_closed = Finished()

        # Wait, until we have the QFileDialog object
        tries = 0
        while (
            file_dialog is None
            or not isinstance(file_dialog, QFileDialog)
            and tries < 10
        ):  # See also https://github.com/pytest-dev/pytest-qt/issues/256
            QApplication.processEvents()
            file_dialog = QApplication.activeModalWidget()
            tries += 1

        if file_dialog is not None:
            file_dialog.findChild(QLineEdit, "fileNameEdit").setText(
                value
            )  # = file_dialog.selectFile(value)

            # Search Open/Choose btn and click it
            for btn in file_dialog.findChildren(QPushButton):
                text = btn.text().lower()
                if "open" in text or "choose" in text:
                    with qtbot.waitSignal(
                        message_box_closed.finished, raising=True
                    ) as _:
                        QTimer.singleShot(
                            5, lambda: closeMessagebox(message_box_closed)
                        )
                        qtbot.mouseClick(btn, Qt.MouseButton.LeftButton)

            file_dialog.close()

    QTimer.singleShot(5, selectFile)
    widget.browse()

    assert widget.get_widget_value() == expected


@pytest.mark.parametrize(
    ("click_attrs", "value", "add_children", "remove_children"),
    [
        (ClickAttrs.nvalue_widget(), "", 0, 0),
        (ClickAttrs.nvalue_widget(), "test", 2, 0),
        (ClickAttrs.nvalue_widget(), "test2", 3, 2),
        (ClickAttrs.nvalue_widget(), "test3", 5, 5),
        # With default
        (ClickAttrs.nvalue_widget(default=["A", "B"]), "test3", 5, 5),
    ],
)
def test_nvaluewidget_add_remove_children(
    click_attrs: dict, value: str, add_children: int, remove_children: int
):
    param = click.Option(param_decls=["--p"], **click_attrs)
    cli = click.Command("cli", params=[param])

    control = clickqt.qtgui_from_click(cli)
    widget: clickqt.widgets.NValueWidget = control.widget_registry[cli.name][param.name]

    for _ in range(add_children):
        widget.add_pair(value)

    amount_children = add_children + (
        len(param.default) if param.default is not None else 0
    )

    assert len(widget.children) == amount_children
    assert (
        sum(1 for w in widget.buttondict.values() if w.get_widget_value() == value)
        == add_children
    )

    for _ in range(remove_children):
        widget.remove_button_pair(list(widget.buttondict.keys())[0])

    assert len(widget.children) == amount_children - remove_children



from __future__ import annotations

import typing as t
import click
from PySide6.QtWidgets import QComboBox

from clickqt.widgets.basewidget import ComboBoxBase, BaseWidget
from clickqt.widgets.core.QCheckableCombobox import QCheckableComboBox


class ComboBox(ComboBoxBase):
    """Represents a click.types.Choice object.

    :param otype: The type which specifies the clickqt widget type. This type may be different compared to **param**.type when dealing with click.types.CompositeParamType-objects
    :param param: The parameter from which **otype** came from
    :param kwargs: Additionally parameters ('parent', 'widgetsource', 'com', 'label') needed for
                    :class:`~clickqt.widgets.basewidget.MultiWidget`- / :class:`~clickqt.widgets.confirmationwidget.ConfirmationWidget`-widgets
    """

    widget_type = QComboBox  #: The Qt-type of this widget.

    def __init__(self, otype: click.ParamType, param: click.Parameter, **kwargs):
        super().__init__(otype, param, **kwargs)

        if (
            self.parent_widget is None
            and (default := BaseWidget.get_param_default(param, None)) is not None
        ):
            self.set_value(default)

    def set_value(self, value: t.Any):
        self.widget.setCurrentText(
            str(
                self.type.convert(
                    str(value), self.click_command, click.Context(self.click_command)
                )
            )
        )

    def add_items(self, items: t.Iterable[str]):
        self.widget.addItems(items)

    def get_widget_value(self) -> str:
        return self.widget.currentText()


class CheckableComboBox(ComboBoxBase):
    """Represents a multiple click.types.Choice object.

    :param otype: The type which specifies the clickqt widget type. This type may be different compared to **param**.type when dealing with click.types.CompositeParamType-objects
    :param param: The parameter from which **otype** came from
    :param kwargs: Additionally parameters ('parent', 'widgetsource', 'com', 'label') needed for
                    :class:`~clickqt.widgets.basewidget.MultiWidget`- / :class:`~clickqt.widgets.confirmationwidget.ConfirmationWidget`-widgets
    """

    widget_type = QCheckableComboBox  #: The Qt-type of this widget.

    def __init__(self, otype: click.ParamType, param: click.Parameter, **kwargs):
        super().__init__(otype, param, **kwargs)

        assert param.multiple, "'param.multiple' should be True"

        if self.parent_widget is None:
            self.set_value(BaseWidget.get_param_default(param, []))

    def set_value(self, value: t.Iterable[t.Any]):
        check_values: list[str] = []
        for v in value:
            check_values.append(
                str(
                    self.type.convert(
                        str(v), self.click_command, click.Context(self.click_command)
                    )
                )
            )

        self.widget.checkItems(check_values)

    def add_items(self, items: t.Iterable[str]):
        self.widget.addItems(items)

    def get_widget_value(self) -> t.Iterable[str]:
        return self.widget.getData()



"""
Tests the external use of clickqt.
"""
from __future__ import annotations

import pytest
import click
from clickqt.__main__ import get_command_from_entrypoint, get_command_from_path


def test_clickqt_external():
    """
    Try some good/bad imports and check if they are recognized as commands/garbage.
    """
    # test with endpoint
    with pytest.raises(ImportError):
        get_command_from_entrypoint("example")
    with pytest.raises(TypeError):
        get_command_from_entrypoint("example_gui")
    assert isinstance(get_command_from_entrypoint("example_cli"), click.Command)

    # test with file
    with pytest.raises(ImportError):
        get_command_from_path("example/example/afwizard.py", "ma")
    with pytest.raises(FileNotFoundError):
        get_command_from_path("example/example/--_--_-_.py", "main")
    with pytest.raises(TypeError):
        get_command_from_path(
            "example/example/afwizard.py", "validate_spatial_reference"
        )
    assert isinstance(
        get_command_from_path("example/example/afwizard.py", "main"), click.Command
    )



from __future__ import annotations

import typing as t

import click
import pytest
from PySide6.QtCore import QEvent

import clickqt.widgets
from tests.testutils import ClickAttrs, raise_


def evaluate(
    clickqt_widget: clickqt.widgets.BaseWidget,
    clickqt_child_widget: clickqt.widgets.BaseWidget,
    invalid_value: t.Any,
    valid_value: t.Any,
):
    value = [invalid_value, valid_value]
    border: list[t.Callable] = [
        lambda widget: f"QWidget#{clickqt_widget.widget_name}{{ border: 1px solid red }}"
        in widget.styleSheet(),  # red border
        lambda widget: f"QWidget#{clickqt_widget.widget_name}{{ }}"
        == widget.styleSheet(),
    ]  # normal border

    for i in range(2):
        clickqt_widget.set_value(value[i])
        clickqt_child_widget.focus_out_validator.eventFilter(
            clickqt_child_widget.widget, QEvent(QEvent.Type.FocusOut)
        )  # widget goes out of focus
        if (
            clickqt_widget == clickqt_child_widget
            and isinstance(clickqt_widget, clickqt.widgets.MultiWidget)
        ) or (
            clickqt_widget != clickqt_child_widget
            and isinstance(
                clickqt_widget,
                (clickqt.widgets.MultiValueWidget, clickqt.widgets.TupleWidget),
            )
        ):  # NValueWidget: Every child can be checked individually
            for child in clickqt_widget.children:
                assert border[i](child.widget)
        else:
            assert border[i](clickqt_child_widget.widget)


@pytest.mark.parametrize(
    ("click_attrs", "invalid_value", "valid_value"),
    [
        (
            ClickAttrs.intfield(
                callback=lambda ctx, param, value: raise_(Exception("..."))
                if value < 5
                else value
            ),
            0,
            10,
        ),
        (ClickAttrs.filepathfield(type_dict={"exists": True}), "invalid_path", "tests"),
        (
            ClickAttrs.tuple_widget(
                types=(str, float),
                callback=lambda ctx, param, value: raise_(Exception("..."))
                if value[0] != "abc"
                else value,
            ),
            ["a", 0],
            ["abc", 2.2],
        ),
        (
            ClickAttrs.multi_value_widget(
                nargs=3,
                type=int,
                callback=lambda ctx, param, value: raise_(Exception("..."))
                if value != (0, 1, 2)
                else value,
            ),
            [0, 0, 0],
            [0, 1, 2],
        ),
        (
            ClickAttrs.nvalue_widget(
                callback=lambda ctx, param, value: raise_(Exception("..."))
                if len(value) == 0
                else value
            ),
            [],
            ["abc"],
        ),
    ],
)
def test_focus_out_validation(click_attrs: dict, invalid_value: t.Any, valid_value: t.Any):
    param = click.Option(param_decls=["--test"], **click_attrs)
    cli = click.Command("cli", params=[param])

    control = clickqt.qtgui_from_click(cli)
    clickqt_widget = control.widget_registry[cli.name][param.name]

    evaluate(clickqt_widget, clickqt_widget, invalid_value, valid_value)


@pytest.mark.parametrize(
    ("click_attrs", "invalid_value", "valid_value"),
    [
        (
            ClickAttrs.tuple_widget(types=(str, click.types.Path(exists=True))),
            ["a", "invalid_path"],
            ["abc", "tests"],
        ),
        (
            ClickAttrs.multi_value_widget(nargs=2, type=click.types.Path(exists=True)),
            ["invalid_path", "invalid_path"],
            ["tests", "tests"],
        ),  # Both children have to be valid
        (
            ClickAttrs.nvalue_widget(type=click.types.Path(exists=True)),
            ["invalid_path", "invalid_path"],
            ["invalid_path", "tests"],
        ),  # We check only children[1], so children[0] can be still invalid
    ],
)
def test_focus_out_validation_child(
    click_attrs: dict, invalid_value: t.Any, valid_value: t.Any
):
    param = click.Option(param_decls=["--test"], **click_attrs)
    cli = click.Command("cli", params=[param])

    control = clickqt.qtgui_from_click(cli)
    clickqt_widget = control.widget_registry[cli.name][param.name]

    clickqt_widget.set_value(
        invalid_value
    )  # Create the children for the NValueWidget-object

    evaluate(
        clickqt_widget, list(clickqt_widget.children)[1], invalid_value, valid_value
    )  # We check the first child



from __future__ import annotations

import click
from PySide6.QtWidgets import QLineEdit
from PySide6.QtGui import QIcon, QAction

from clickqt.widgets.textfield import TextField


class PasswordField(TextField):
    """Represents a click.types.StringParamType-object with hide_input==True.
    The input will be hidden by default, but can be made visible by clicking on the "eye-show" icon.

    :param otype: The type which specifies the clickqt widget type. This type may be different compared to **param**.type when dealing with click.types.CompositeParamType-objects
    :param param: The parameter from which **otype** came from
    :param kwargs: Additionally parameters ('parent', 'widgetsource', 'com', 'label') needed for
                    :class:`~clickqt.widgets.basewidget.MultiWidget`- / :class:`~clickqt.widgets.confirmationwidget.ConfirmationWidget`-widgets
    """

    widget_type = QLineEdit  #: The Qt-type of this widget.

    def __init__(self, otype: click.ParamType, param: click.Parameter, **kwargs):
        super().__init__(otype, param, **kwargs)

        assert (
            hasattr(param, "hide_input") and param.hide_input
        ), "'param.hide_input' should be True"

        self.icon_text = (
            (QIcon("clickqt\\images\\eye-show.png"), "Show password"),
            (QIcon("clickqt\\images\\eye-hide.png"), "Hide password"),
        )
        self.show_hide_action = QAction(
            icon=self.icon_text[0][0], text=self.icon_text[0][1]
        )
        self.widget.setEchoMode(QLineEdit.EchoMode.Password)
        self.widget.addAction(
            self.show_hide_action, QLineEdit.ActionPosition.TrailingPosition
        )
        self.show_hide_action.setCheckable(True)

        def show_password(show):
            self.widget.setEchoMode(
                QLineEdit.EchoMode.Normal if show else QLineEdit.EchoMode.Password
            )
            self.show_hide_action.setIcon(self.icon_text[show][0])
            self.show_hide_action.setText(self.icon_text[show][1])

        self.show_hide_action.toggled.connect(show_password)



""" Contains the checkbox widget """
from __future__ import annotations

import typing as t

from PySide6.QtWidgets import QCheckBox
import click

from clickqt.widgets.basewidget import BaseWidget


class CheckBox(BaseWidget):
    """Represents a click.types.BoolParamType object.

    :param otype: The type which specifies the clickqt widget type.
        This type may be different compared to **param**.type when dealing with click.types.CompositeParamType-objects
    :param param: The parameter from which **otype** came from
    :param kwargs: Additionally parameters ('parent', 'widgetsource', 'com', 'label') needed for
        :class:`~clickqt.widgets.basewidget.MultiWidget`- /
        :class:`~clickqt.widgets.confirmationwidget.ConfirmationWidget`- widgets
    """

    widget_type = QCheckBox  #: The Qt-type of this widget.

    def __init__(self, otype: click.ParamType, param: click.Parameter, **kwargs):
        super().__init__(otype, param, **kwargs)

        assert isinstance(
            otype, type(click.BOOL)
        ), f"'otype' must be of type '{type(click.BOOL)}', but is '{type(otype)}'."

        self.widget.setText("Enable")

        if self.parent_widget is None:
            self.set_value(BaseWidget.get_param_default(param, False))

    def set_value(self, value: t.Any):
        self.widget.setChecked(
            bool(
                self.type.convert(
                    str(value), self.click_command, click.Context(self.click_command)
                )
            )
        )

    def get_widget_value(self) -> bool:
        return self.widget.isChecked()



"""
Contains tests for widget default values
"""
from __future__ import annotations

import datetime
import typing as t

import click
import pytest

from tests.testutils import ClickAttrs
import clickqt.widgets


@pytest.mark.parametrize(
    ("click_attrs", "default", "expected"),
    [
        (ClickAttrs.checkbox(), True, True),
        (ClickAttrs.checkbox(), 1, True),
        (ClickAttrs.checkbox(), "yes", True),
        (ClickAttrs.checkbox(), False, False),
        (ClickAttrs.checkbox(), 0, False),
        (ClickAttrs.checkbox(), "no", False),
        (ClickAttrs.intfield(), 12, 12),
        (ClickAttrs.intfield(), -1322, -1322),
        (ClickAttrs.intfield(), "-31", -31),
        (ClickAttrs.realfield(), 152.31, 152.31),
        (ClickAttrs.realfield(), -123.2, -123.2),
        (ClickAttrs.realfield(), "1.23", 1.23),
        (ClickAttrs.intrange(), "1", 1),
        (ClickAttrs.intrange(maxval=2, clamp=True), 5, 2),
        (ClickAttrs.intrange(minval=2, clamp=True), -1, 2),
        (ClickAttrs.floatrange(), "1.23", 1.23),
        (ClickAttrs.floatrange(maxval=2.5, clamp=True), 5.2, 2.5),
        (ClickAttrs.floatrange(minval=2.5, clamp=True), -1, 2.5),
        (ClickAttrs.textfield(), "test123", "test123"),
        (ClickAttrs.textfield(), 12.2, "12.2"),
        (ClickAttrs.passwordfield(), "abc", "abc"),
        (ClickAttrs.confirmation_widget(), "test321", "test321"),
        (ClickAttrs.messagebox(prompt="Test"), True, True),
        (ClickAttrs.messagebox(prompt="Test"), "y", True),
        (ClickAttrs.messagebox(prompt="Test"), "on", True),
        (ClickAttrs.messagebox(prompt="Test"), False, False),
        (ClickAttrs.messagebox(prompt="Test"), "n", False),
        (ClickAttrs.messagebox(prompt="Test"), "off", False),
        (ClickAttrs.combobox(choices=["A", "B", "C"]), "B", "B"),
        (ClickAttrs.combobox(choices=["A", "B", "C"], case_sensitive=False), "b", "B"),
        (
            ClickAttrs.checkable_combobox(choices=["A", "B", "C"]),
            ["B", "C"],
            ["B", "C"],
        ),
        (
            ClickAttrs.checkable_combobox(
                choices=["A", "B", "C"], case_sensitive=False
            ),
            ["a", "c"],
            ["A", "C"],
        ),
        (
            ClickAttrs.datetime(),
            "2023-06-23",
            datetime.datetime.strptime("2023-06-23 00:00:00", "%Y-%m-%d %H:%M:%S"),
        ),  # Use the default formats
        (
            ClickAttrs.datetime(),
            "2023-06-23 15:14:20",
            datetime.datetime.strptime("2023-06-23 15:14:20", "%Y-%m-%d %H:%M:%S"),
        ),
        (
            ClickAttrs.datetime(formats=["%d-%m-%Y"]),
            "23-06-2023",
            datetime.datetime.strptime("23-06-2023", "%d-%m-%Y"),
        ),
        (
            ClickAttrs.datetime(formats=["%d-%m-%Y %H:%M:%S", "%d-%m-%Y"]),
            "23-06-2023 12:30:20",
            datetime.datetime.strptime("23-06-2023", "%d-%m-%Y"),
        ),  # Default: Use the last format
        (ClickAttrs.filefield(), "test.abc", "test.abc"),
        (ClickAttrs.filefield(), 123.2, "123.2"),
        (ClickAttrs.filepathfield(), ".", "."),
        (ClickAttrs.filepathfield(), 92.3, "92.3"),
        (ClickAttrs.tuple_widget(types=(str, int)), ["s", "1"], ["s", 1]),
        (
            ClickAttrs.tuple_widget(types=(str, int, float)),
            ["t", 1, "-2."],
            ["t", 1, -2.0],
        ),
        (ClickAttrs.multi_value_widget(nargs=3), ["a", "b", "c"], ["a", "b", "c"]),
        (
            ClickAttrs.multi_value_widget(nargs=3, type=float),
            [1.2, "-3.5", -2],
            [1.2, -3.5, -2],
        ),
        (ClickAttrs.nvalue_widget(type=int), [1, -2, 5], [1, -2, 5]),
        (
            ClickAttrs.nvalue_widget(type=(str, float)),
            [["a", 12.2], ["b", -873.21]],
            [["a", 12.2], ["b", -873.21]],
        ),
        (
            ClickAttrs.nvalue_widget(type=(str, (int, str))),
            [["a", [12, "b"]], ["c", [-1, "d"]]],
            [["a", [12, "b"]], ["c", [-1, "d"]]],
        ),
        (
            ClickAttrs.nvalue_widget(type=(str, (int, float))),
            lambda: [["a", [1, 2.1]], ["b", [3, -4.2]]],
            [["a", [1, 2.1]], ["b", [3, -4.2]]],
        ),  # callable as default
    ],
)
def test_set_default(click_attrs: dict, default: t.Any, expected: t.Any):
    """Shorthand to check that the correct default is set"""
    param = click.Option(param_decls=["--test"], default=default, **click_attrs)
    cli = click.Command("cli", params=[param])

    control = clickqt.qtgui_from_click(cli)
    assert control.widget_registry[cli.name][param.name].get_widget_value() == expected


@pytest.mark.parametrize(
    ("click_attrs", "default", "expected"),
    [
        (ClickAttrs.checkbox(), 12, "'12' is not a valid boolean."),
        (ClickAttrs.checkbox(), "ok", "'ok' is not a valid boolean."),
        (ClickAttrs.checkbox(), "-1.0", "'-1.0' is not a valid boolean."),
        (ClickAttrs.intfield(), -2.12, "'-2.12' is not a valid integer."),
        (ClickAttrs.intfield(), "True", "'True' is not a valid integer."),
        (ClickAttrs.intfield(), "-1.0", "'-1.0' is not a valid integer."),
        (ClickAttrs.realfield(), "no", "'no' is not a valid float."),
        (ClickAttrs.intrange(minval=0), -1, "-1 is not in the range x>=0."),
        (ClickAttrs.intrange(minval=1, min_open=True), 1, "1 is not in the range x>1."),
        (ClickAttrs.intrange(maxval=1), 2, "2 is not in the range x<=1."),
        (ClickAttrs.intrange(maxval=1, max_open=True), 1, "1 is not in the range x<1."),
        (ClickAttrs.intrange(), -2.12, "'-2.12' is not a valid integer range."),
        (ClickAttrs.floatrange(minval=0.5), -1.2, "-1.2 is not in the range x>=0.5."),
        (
            ClickAttrs.floatrange(minval=1.2, min_open=True),
            1.2,
            "1.2 is not in the range x>1.2.",
        ),
        (ClickAttrs.floatrange(maxval=1.0), 2.2, "2.2 is not in the range x<=1.0."),
        (
            ClickAttrs.floatrange(maxval=1.0, max_open=True),
            1,
            "1.0 is not in the range x<1.0.",
        ),
        (ClickAttrs.floatrange(), "y", "'y' is not a valid float range."),
        (ClickAttrs.combobox(choices=["A", "B"]), "b", "'b' is not one of 'A', 'B'."),
        (
            ClickAttrs.combobox(choices=["A", "B"], case_sensitive=False),
            "C",
            "'C' is not one of 'A', 'B'.",
        ),
        (
            ClickAttrs.checkable_combobox(choices=["A", "B", "C"]),
            ["b", "C"],
            "'b' is not one of 'A', 'B', 'C'.",
        ),
        (
            ClickAttrs.checkable_combobox(
                choices=["A", "B", "C"], case_sensitive=False
            ),
            ["a", "t", "c"],
            "'t' is not one of 'A', 'B', 'C'.",
        ),
        (
            ClickAttrs.datetime(),
            "23-06-2023",
            "'23-06-2023' does not match the formats '%Y-%m-%d', '%Y-%m-%dT%H:%M:%S', '%Y-%m-%d %H:%M:%S'.",
        ),
        (
            ClickAttrs.datetime(formats=["%d-%m-%Y", "%d-%m-%Y %H:%M:%S"]),
            "23-066-2023",
            "'23-066-2023' does not match the formats '%d-%m-%Y', '%d-%m-%Y %H:%M:%S'.",
        ),  # User defined formats
        (
            ClickAttrs.multi_value_widget(nargs=3, type=int),
            ["1", 2, 3.2],
            "'3.2' is not a valid integer.",
        ),
        (
            ClickAttrs.multi_value_widget(nargs=2, type=float),
            ["yes", "y"],
            "'yes' is not a valid float.",
        ),  # First wrong value fails the test
        (
            ClickAttrs.tuple_widget(types=(str, int)),
            ["s", 12.3],
            "'12.3' is not a valid integer.",
        ),
        (
            ClickAttrs.nvalue_widget(type=(str, float)),
            [["a", "set"], ["b", "t"]],
            "'set' is not a valid float.",
        ),  # First wrong value fails the test
    ],
)
def test_set_default_fail(click_attrs: dict, default: t.Any, expected: t.Any):
    """Shorthand to check that the correct error is returned"""
    param = click.Option(param_decls=["--test"], default=default, **click_attrs)
    cli = click.Command("cli", params=[param])

    with pytest.raises(click.exceptions.BadParameter) as exc_info:
        # pylint: disable=w0106
        clickqt.qtgui_from_click(cli).widget_registry[cli.name][param.name]

    assert expected == exc_info.value.message



from __future__ import annotations

import inspect
from collections import defaultdict
import typing as t

import click

from PySide6.QtWidgets import QApplication
from PySide6.QtCore import QThread


def raise_(ex: Exception):
    raise ex


def wait_process_Events(ms: int, x_times: int = 3):
    for _ in range(x_times):
        QApplication.processEvents()
        QThread.msleep(ms)


# Credits to https://stackoverflow.com/questions/15788725/how-to-determine-the-closest-common-ancestor-class
def clcoancl(*cls_list):
    mros = [list(inspect.getmro(cls)) for cls in cls_list]
    track = defaultdict(int)
    while mros:
        for mro in mros:
            cur = mro.pop(0)
            track[cur] += 1
            if track[cur] == len(cls_list):
                return cur
            if len(mro) == 0:
                mros.remove(mro)
    return None  # or raise, if that's more appropriate


class ClickAttrs:
    @staticmethod
    def unprocessed(**attrs_dict) -> dict:
        return {"type": click.types.UNPROCESSED, **attrs_dict}

    @staticmethod
    def messagebox(prompt: str, **attrs_dict) -> dict:
        return {
            "type": click.types.BOOL,
            "is_flag": True,
            "prompt": prompt,
            **attrs_dict,
        }

    @staticmethod
    def intfield(**attrs_dict) -> dict:
        return {"type": click.types.INT, **attrs_dict}

    @staticmethod
    def passwordfield(**attrs_dict) -> dict:
        return {"type": click.types.STRING, "hide_input": True, **attrs_dict}

    @staticmethod
    def realfield(**attrs_dict) -> dict:
        return {"type": click.types.FLOAT, **attrs_dict}

    @staticmethod
    def intrange(
        minval: "int|None" = None,
        maxval: "int|None" = None,
        min_open: bool = False,
        max_open: bool = False,
        clamp: bool = False,
        **attrs_dict
    ) -> dict:
        return {
            "type": click.IntRange(
                min=minval,
                max=maxval,
                min_open=min_open,
                max_open=max_open,
                clamp=clamp,
            ),
            **attrs_dict,
        }

    @staticmethod
    def floatrange(
        minval: "float|None" = None,
        maxval: "float|None" = None,
        min_open: bool = False,
        max_open: bool = False,
        clamp: bool = False,
        **attrs_dict
    ) -> dict:
        return {
            "type": click.FloatRange(
                min=minval,
                max=maxval,
                min_open=min_open,
                max_open=max_open,
                clamp=clamp,
            ),
            **attrs_dict,
        }

    @staticmethod
    def confirmation_widget(**attrs_dict) -> dict:
        return {"confirmation_prompt": True, **attrs_dict}

    @staticmethod
    def checkbox(**attrs_dict) -> dict:
        return {"type": click.types.BOOL, **attrs_dict}

    @staticmethod
    def combobox(
        choices: t.Sequence[str], case_sensitive: bool = True, **attrs_dict
    ) -> dict:
        return {
            "type": click.types.Choice(choices=choices, case_sensitive=case_sensitive),
            **attrs_dict,
        }

    @staticmethod
    def checkable_combobox(
        choices: t.Sequence[str], case_sensitive: bool = True, **attrs_dict
    ) -> dict:
        return {
            "type": click.types.Choice(choices=choices, case_sensitive=case_sensitive),
            "multiple": True,
            **attrs_dict,
        }

    @staticmethod
    def filefield(type_dict: dict = {}, **attrs_dict) -> dict:
        return {"type": click.types.File(**type_dict), **attrs_dict}

    @staticmethod
    def filepathfield(type_dict: dict = {}, **attrs_dict) -> dict:
        return {"type": click.types.Path(**type_dict), **attrs_dict}

    @staticmethod
    def datetime(formats: t.Optional[t.Sequence[str]] = None, **attrs_dict) -> dict:
        return {"type": click.types.DateTime(formats), **attrs_dict}

    @staticmethod
    def tuple_widget(
        types: t.Sequence[t.Union[t.Type[t.Any], click.ParamType]], **attrs_dict
    ) -> dict:
        return {"type": click.types.Tuple(types), **attrs_dict}

    @staticmethod
    def multi_value_widget(nargs: int, **attrs_dict) -> dict:
        assert nargs > 1
        assert attrs_dict.get("type") is None or not isinstance(
            attrs_dict["type"], tuple
        )
        return {"nargs": nargs, **attrs_dict}

    @staticmethod
    def nvalue_widget(**attrs_dict) -> dict:
        assert attrs_dict.get("type") is None or attrs_dict.get("type") is not bool
        return {"multiple": True, **attrs_dict}

    @staticmethod
    def uuid(**attrs_dict) -> dict:
        return {"type": click.types.UUID, **attrs_dict}

    @staticmethod
    def textfield(**attrs_dict) -> dict:
        return {"type": click.types.STRING, **attrs_dict}



from __future__ import annotations

from PySide6.QtWidgets import (
    QWidget,
    QTreeView,
    QLineEdit,
    QPushButton,
    QFileDialog,
    QDialog,
)
from PySide6.QtCore import QObject, QEvent, SIGNAL, QDir, Slot


class QPathDialog(QFileDialog):
    """A file dialog that accepts a single file or a single directory.

    :param parent: The parent Qt-widget, defaults to None
    :param directory: The directory that will be displayed when opening the dialog, defaults to the current path
    :param exist: Specifies, whether the file / directory has to exist
    """

    def __init__(
        self,
        parent: "QWidget | None" = None,
        directory: str = QDir.currentPath(),
        exist: bool = True,
    ):
        super().__init__(parent, directory=directory)

        self.exist = exist
        self.setOption(QFileDialog.Option.DontUseNativeDialog, True)
        self.setFileMode(QFileDialog.FileMode.Directory)
        self.open_button: QPushButton = None
        # self.list_view: QListView = self.findChild(QListView, "listView")
        # if self.list_view:
        #    self.list_view.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
        self.tree_view: QTreeView = self.findChild(QTreeView)
        # if self.tree_view:
        #    self.tree_view.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
        self.selected_path: str = ""
        self.base_dir: str = ""

        for btn in self.findChildren(QPushButton):
            text = btn.text().lower()
            if "open" in text or "choose" in text:
                self.open_button = btn
                break

        if self.open_button:
            self.open_button.installEventFilter(self)
            self.disconnect(self.open_button, SIGNAL("clicked()"))
            self.open_button.clicked.connect(self.openClicked)

    def selectedPath(self) -> str:
        """Returns the selected path."""

        return self.selected_path

    def eventFilter(self, watched: QObject, event: QEvent) -> bool:
        btn: QPushButton = watched
        if btn:
            if event.type() == QEvent.Type.EnabledChange and not btn.isEnabled():
                btn.setEnabled(True)
            elif event.type() == QEvent.Type.MouseButtonRelease:
                self.base_dir = (
                    self.directory().absolutePath()
                )  # Save current dir, after changing fileMode it could be changed
                if not self.exist:
                    self.setFileMode(QFileDialog.FileMode.AnyFile)

        return QWidget.eventFilter(self, watched, event)

    @Slot()
    def openClicked(self):
        self.selected_path = ""
        for model_index in self.tree_view.selectionModel().selectedIndexes():
            if model_index.column() == 0:
                self.selected_path = (
                    self.base_dir + QDir.separator() + str(model_index.data())
                )
                self.selected_path.replace("/", QDir.separator())
                break

        if not self.selected_path and not self.exist:
            self.selected_path = (
                self.base_dir
                + QDir.separator()
                + self.findChild(QLineEdit, "fileNameEdit").text()
            )
            self.selected_path.replace("/", QDir.separator())
            QDialog.accept(self)
        elif (
            self.selected_path
        ):  # Don't close if the user selected an invalid file/folder
            QDialog.accept(self)



""" Contains the GUI class. """
from __future__ import annotations

import sys
import click
from PySide6.QtWidgets import (
    QApplication,
    QSplitter,
    QWidget,
    QVBoxLayout,
    QHBoxLayout,
    QPushButton,
    QSizePolicy,
)
from PySide6.QtGui import (
    QColor,
    Qt,
    QPalette,
    QScreen,
)
from clickqt.widgets.multivaluewidget import MultiValueWidget
from clickqt.widgets.basewidget import BaseWidget
from clickqt.widgets.checkbox import CheckBox
from clickqt.widgets.textfield import TextField
from clickqt.widgets.passwordfield import PasswordField
from clickqt.widgets.numericfields import IntField, RealField
from clickqt.widgets.combobox import ComboBox, CheckableComboBox
from clickqt.widgets.datetimeedit import DateTimeEdit
from clickqt.widgets.tuplewidget import TupleWidget
from clickqt.widgets.filepathfield import FilePathField
from clickqt.widgets.filefield import FileField
from clickqt.widgets.nvaluewidget import NValueWidget
from clickqt.widgets.confirmationwidget import ConfirmationWidget
from clickqt.widgets.messagebox import MessageBox
from clickqt.core.output import OutputStream, TerminalOutput


class GUI:
    """
    Responsible for setting up the components for the Qt-GUI,
    which is used to navigate through the different kind of commands and execute them.
    """

    def __init__(self):
        self.window = QWidget()
        self.window.setLayout(QVBoxLayout())
        self.splitter = QSplitter(Qt.Orientation.Vertical)
        self.splitter.setChildrenCollapsible(
            False
        )  # Child widgets can't be resized down to size 0
        self.window.layout().addWidget(self.splitter)

        self.widgets_container: QWidget = None  # Control constructs this Qt-widget

        self.buttons_container = QWidget()
        self.buttons_container.setLayout(QHBoxLayout())
        self.buttons_container.setSizePolicy(
            QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed
        )  # Not resizable in vertical direction
        self.run_button = QPushButton("&Run")  # Shortcut Alt+R
        self.stop_button = QPushButton("&Stop")  # Shortcut Alt+S
        self.stop_button.setEnabled(False)
        self.copy_button = QPushButton("&Copy-To-Clipboard")
        self.buttons_container.layout().addWidget(self.run_button)
        self.buttons_container.layout().addWidget(self.stop_button)
        self.buttons_container.layout().addWidget(self.copy_button)

        self.terminal_output = TerminalOutput()
        self.terminal_output.setReadOnly(True)
        self.terminal_output.setToolTip("Terminal output")
        self.terminal_output.newHtmlMessage.connect(self.terminal_output.writeHtml)

        sys.stdout = OutputStream(
            self.terminal_output, sys.stdout, QPalette().color(QPalette.ColorRole.Text)
        )
        sys.stderr = OutputStream(self.terminal_output, sys.stderr, QColor("red"))

    def __call__(self):
        """Shows the GUI-window"""

        self.window.show()
        QApplication.instance().exec()

    def __del__(self):
        """Resets the default streams"""

        if isinstance(sys.stdout, OutputStream):
            sys.stdout = sys.stdout.stream

        if isinstance(sys.stderr, OutputStream):
            sys.stderr = sys.stderr.stream

    def construct(self):
        """Resize and reposition the window."""
        assert self.widgets_container is not None

        self.splitter.addWidget(self.widgets_container)
        self.splitter.addWidget(self.buttons_container)
        self.splitter.addWidget(self.terminal_output)

        size_hint = self.window.sizeHint()
        self.window.resize(
            1.5 * size_hint.width(), size_hint.height()
        )  # Enlarge window width

        center = QScreen.availableGeometry(QApplication.primaryScreen()).center()
        geo = self.window.geometry()
        geo.moveCenter(center)
        self.window.move(geo.topLeft())

    def create_widget(  # pylint: disable=no-self-use
        self, otype: click.ParamType, param: click.Parameter, **kwargs
    ) -> BaseWidget:
        """
        Creates the clickqt widget object of the correct widget class determined by the **otype**
        and returns it.

        :param otype: The type which specifies the clickqt widget type.
            This type may be differ from the **param**.type when dealing with
            click.types.CompositeParamType-objects

        :param param: The parameter from which **otype** came from

        :param kwargs: Additional parameters ('widgetsource', 'parent', 'com')
            needed for :class:`~clickqt.widgets.basewidget.MultiWidget`-widgets
        """

        typedict = {
            click.types.BoolParamType: MessageBox
            if hasattr(param, "is_flag")
            and param.is_flag
            and hasattr(param, "prompt")
            and param.prompt
            else CheckBox,
            click.types.IntParamType: IntField,
            click.types.FloatParamType: RealField,
            click.types.StringParamType: PasswordField
            if hasattr(param, "hide_input") and param.hide_input
            else TextField,
            click.types.UUIDParameterType: TextField,
            click.types.UnprocessedParamType: TextField,
            click.types.DateTime: DateTimeEdit,
            click.types.Tuple: TupleWidget,
            click.types.Choice: ComboBox,
            click.types.Path: FilePathField,
            click.types.File: FileField,
        }

        def get_multiarg_version(otype: click.ParamType):
            if isinstance(otype, click.types.Choice):
                return CheckableComboBox
            return NValueWidget

        if hasattr(param, "confirmation_prompt") and param.confirmation_prompt:
            return ConfirmationWidget(otype, param, **kwargs)
        if param.multiple:
            return get_multiarg_version(otype)(otype, param, **kwargs)
        if param.nargs > 1:
            if isinstance(otype, click.types.Tuple):
                return TupleWidget(otype, param, **kwargs)
            return MultiValueWidget(otype, param, **kwargs)

        for t, widgetclass in typedict.items():
            if isinstance(otype, t):
                return widgetclass(otype, param, **kwargs)

        return TextField(otype, param, **kwargs)  # Custom types are mapped to TextField



"""
Contains the entry point for clickqt, called clickqtfy, allowing external use.
"""
from __future__ import annotations

import sys
from importlib import util, metadata

import click
from clickqt.core.core import qtgui_from_click


@click.command("clickqtfy")
@click.argument("entrypoint")
@click.argument("funcname", default=None, required=False)
def clickqtfy(entrypoint, funcname):
    """
    Generate a GUI for an entry point or a file + click.command combinaiton.

    ENTRYPOINT: Name of an installed entry point or a file path.\n
    FUNCNAME: Name of the click.command inside the file at ENTRYPOINT.\n
    If FUNCNAME is provided, ENTRYPOINT is interpreted as a file. Otherwise, as an entry point.
    """
    if funcname:
        fileparam = click.types.File()
        fileparam.convert(entrypoint, None, None)
        control = qtgui_from_click(get_command_from_path(entrypoint, funcname))
        control.set_ep_or_path(entrypoint)
        control.set_is_ep(False)
    else:
        control = qtgui_from_click(get_command_from_entrypoint(entrypoint))
        control.set_ep_or_path(entrypoint)
        control.set_is_ep(True)
    
    return control()


def get_command_from_entrypoint(epname: str) -> click.Command:
    """
    Returns the click.Command specified by `epname`.
    If `epname` is not a click.Command, raises `ImportError`.
    """
    eps = get_entrypoints_from_name(epname)
    if len(eps) == 0:
        raise ImportError(f"No entry point named '{epname}' found.")
    if len(eps) > 1:
        concateps = "\n".join([ep.name for ep in eps])
        raise ImportError(
            f"No entry point named '{epname}' found. Similar ones:\n{concateps}"
        )
    return validate_entrypoint(eps[0].load())


def get_entrypoints_from_name(epname: str) -> list[metadata.EntryPoint]:
    """
    Returns the entrypoints that include `epname` in their name.
    """
    grouped_eps = metadata.entry_points()
    candidates: list[metadata.EntryPoint] = []
    for group in grouped_eps.values():
        for entrypoint in group:
            if entrypoint.name == epname:
                return [entrypoint]
            if epname in entrypoint.name or epname in entrypoint.value:
                candidates.append(entrypoint)
    return candidates


def get_command_from_path(eppath: str, epname: str) -> click.Command:
    """
    Returns the entrypoint given by the file path and the function name,
    or raises `ImportError` if the endpoint is not a `click.Command`.
    """
    modulename = "clickqtfy.imported_module"
    spec = util.spec_from_file_location(modulename, eppath)
    module = util.module_from_spec(spec)
    sys.modules[modulename] = module
    spec.loader.exec_module(module)
    entrypoint = getattr(module, epname, None)
    if entrypoint is None:
        raise ImportError(
            f"Module '{spec.origin}' does not contain the entry point '{epname}'."
        )
    return validate_entrypoint(entrypoint)


def validate_entrypoint(entrypoint):
    """
    Raise a `TypeError` if a provided function is not a `click.Command`.
    """
    if not isinstance(entrypoint, click.Command):
        raise TypeError(f"Entry point '{entrypoint}' is not a 'click.Command'.")
    return entrypoint



import typing as t
import pytest
import click
from PySide6.QtWidgets import QApplication
from PySide6.QtGui import QClipboard
import clickqt.widgets
from tests.testutils import ClickAttrs


def prepare_execution(cmd: click.Command, cmd_group_name: click.Group):
    return cmd_group_name.name + ":" + cmd.name


@pytest.mark.parametrize(
    ("click_attrs", "value", "expected_output"),
    [
        (ClickAttrs.intfield(), 12, "main  --p 12"),
        (ClickAttrs.textfield(), "test", "main  --p test"),
        (ClickAttrs.realfield(), 0.8, "main  --p 0.8"),
        (ClickAttrs.passwordfield(), "abc", "main  --p abc"),
        (ClickAttrs.checkbox(), True, "main  --p True"),
        (ClickAttrs.checkbox(), False, "main  --p False"),
        (ClickAttrs.intrange(maxval=2, clamp=True), 5, "main  --p 2"),
        (ClickAttrs.floatrange(maxval=2.05, clamp=True), 5, "main  --p 2.05"),
        (
            ClickAttrs.combobox(
                choices=["A", "B", "C"], case_sensitive=False, confirmation_prompt=True
            ),
            "B",
            "main  --p=B",
        ),
        (
            ClickAttrs.combobox(choices=["A", "B", "C"], case_sensitive=False),
            "B",
            "main  --p=B",
        ),
        (
            ClickAttrs.checkable_combobox(choices=["A", "B", "C"]),
            ["B", "C"],
            "main  --p=B --p=C",
        ),
        (ClickAttrs.checkable_combobox(choices=["A", "B", "C"]), ["A"], "main  --p=A"),
        (
            ClickAttrs.checkable_combobox(choices=["A", "B", "C"]),
            ["A", "B", "C"],
            "main  --p=A --p=B --p=C",
        ),
        (
            ClickAttrs.tuple_widget(types=(str, int, float)),
            ("t", 1, -2.0),
            "main  --p  t  1  -2.0",
        ),
        (
            ClickAttrs.nvalue_widget(type=(str, int)),
            [["a", 12], ["b", 11]],
            "main  --p   a   12  --p   b   11 ",
        ),
        (
            (
                ClickAttrs.multi_value_widget(nargs=2),
                ["foo", "bar"],
                "main  --p  foo  bar",
            )
        ),
        (
            ClickAttrs.multi_value_widget(nargs=2, default=["A", "B"]),
            ["A", "C"],
            "main  --p  A  C",
        ),
        (
            ClickAttrs.multi_value_widget(nargs=2, default=["A", "B"]),
            [" ", " "],
            "main  --p      ",
        ),
        (
            ClickAttrs.nvalue_widget(type=(click.types.File(), int)),
            [[".gitignore", 12], ["setup.py", -1]],
            "main  --p   .gitignore   12  --p   setup.py   -1 ",
        ),
    ],
)
def test_command_with_ep(click_attrs: dict, value: t.Any, expected_output: str):
    param = click.Option(param_decls=["--p"], **click_attrs)
    cli = click.Command("cli", params=[param])
    control = clickqt.qtgui_from_click(cli)
    control.set_ep_or_path("main")
    control.set_is_ep(True)
    widget = control.widget_registry[cli.name][param.name]
    widget.set_value(value)

    control.construct_command_string()

    assert control.is_ep is True
    assert control.ep_or_path == "main"
    assert control.cmd == cli

    # Simulate clipboard behavior using QApplication.clipboard()
    clipboard = QApplication.clipboard()
    print(clipboard.text(QClipboard.Clipboard))
    assert clipboard.text(QClipboard.Clipboard) == expected_output


@pytest.mark.parametrize(
    ("click_attrs", "value", "expected_output"),
    [
        (ClickAttrs.intfield(), 12, "python example/example/main.py  --p 12"),
        (ClickAttrs.textfield(), "test", "python example/example/main.py  --p test"),
        (ClickAttrs.realfield(), 0.8, "python example/example/main.py  --p 0.8"),
        (ClickAttrs.passwordfield(), "abc", "python example/example/main.py  --p abc"),
        (ClickAttrs.checkbox(), True, "python example/example/main.py  --p True"),
        (ClickAttrs.checkbox(), False, "python example/example/main.py  --p False"),
        (
            ClickAttrs.intrange(maxval=2, clamp=True),
            5,
            "python example/example/main.py  --p 2",
        ),
        (
            ClickAttrs.floatrange(maxval=2.05, clamp=True),
            5,
            "python example/example/main.py  --p 2.05",
        ),
        (
            ClickAttrs.combobox(
                choices=["A", "B", "C"], case_sensitive=False, confirmation_prompt=True
            ),
            "B",
            "python example/example/main.py  --p=B",
        ),
        (
            ClickAttrs.combobox(choices=["A", "B", "C"], case_sensitive=False),
            "B",
            "python example/example/main.py  --p=B",
        ),
        (
            ClickAttrs.checkable_combobox(choices=["A", "B", "C"]),
            ["B", "C"],
            "python example/example/main.py  --p=B --p=C",
        ),
        (
            ClickAttrs.checkable_combobox(choices=["A", "B", "C"]),
            ["A"],
            "python example/example/main.py  --p=A",
        ),
        (
            ClickAttrs.checkable_combobox(choices=["A", "B", "C"]),
            ["A", "B", "C"],
            "python example/example/main.py  --p=A --p=B --p=C",
        ),
        (
            ClickAttrs.tuple_widget(types=(str, int, float)),
            ("t", 1, -2.0),
            "python example/example/main.py  --p  t  1  -2.0",
        ),
        (
            ClickAttrs.nvalue_widget(type=(str, int)),
            [["a", 12], ["b", 11]],
            "python example/example/main.py  --p   a   12  --p   b   11 ",
        ),
        (
            (
                ClickAttrs.multi_value_widget(nargs=2),
                ["foo", "bar"],
                "python example/example/main.py  --p  foo  bar",
            )
        ),
        (
            ClickAttrs.multi_value_widget(nargs=2, default=["A", "B"]),
            ["A", "C"],
            "python example/example/main.py  --p  A  C",
        ),
        (
            ClickAttrs.multi_value_widget(nargs=2, default=["A", "B"]),
            [" ", " "],
            "python example/example/main.py  --p      ",
        ),
        (
            ClickAttrs.nvalue_widget(type=(click.types.File(), int)),
            [[".gitignore", 12], ["setup.py", -1]],
            "python example/example/main.py  --p   .gitignore   12  --p   setup.py   -1 ",
        ),
    ],
)
def test_construct_cmd_string_file(
    click_attrs: dict, value: t.Any, expected_output: str
):
    param = click.Option(param_decls=["--p"], **click_attrs)
    cli = click.Command("cli", params=[param])
    control = clickqt.qtgui_from_click(cli)
    control.set_ep_or_path("example/example/main.py")
    control.set_is_ep(False)
    widget = control.widget_registry[cli.name][param.name]
    widget.set_value(value)

    control.construct_command_string()

    assert control.is_ep is False
    assert control.ep_or_path == "example/example/main.py"
    assert control.cmd == cli

    # Simulate clipboard behavior using QApplication.clipboard()
    clipboard = QApplication.clipboard()
    print(clipboard.text(QClipboard.Clipboard))
    assert clipboard.text(QClipboard.Clipboard) == expected_output


@pytest.mark.parametrize(
    ("click_attrs", "value", "expected_output"),
    [
        (ClickAttrs.intfield(), 12, "main  cmd --p 12"),
        (ClickAttrs.textfield(), "test", "main  cmd --p test"),
        (ClickAttrs.realfield(), 0.8, "main  cmd --p 0.8"),
        (ClickAttrs.passwordfield(), "abc", "main  cmd --p abc"),
        (ClickAttrs.checkbox(), True, "main  cmd --p True"),
        (ClickAttrs.checkbox(), False, "main  cmd --p False"),
        (ClickAttrs.intrange(maxval=2, clamp=True), 5, "main  cmd --p 2"),
        (ClickAttrs.floatrange(maxval=2.05, clamp=True), 5, "main  cmd --p 2.05"),
        (
            ClickAttrs.combobox(
                choices=["A", "B", "C"], case_sensitive=False, confirmation_prompt=True
            ),
            "B",
            "main  cmd --p=B",
        ),
        (
            ClickAttrs.combobox(choices=["A", "B", "C"], case_sensitive=False),
            "B",
            "main  cmd --p=B",
        ),
        (
            ClickAttrs.checkable_combobox(choices=["A", "B", "C"]),
            ["B", "C"],
            "main  cmd --p=B --p=C",
        ),
        (
            ClickAttrs.checkable_combobox(choices=["A", "B", "C"]),
            ["A"],
            "main  cmd --p=A",
        ),
        (
            ClickAttrs.checkable_combobox(choices=["A", "B", "C"]),
            ["A", "B", "C"],
            "main  cmd --p=A --p=B --p=C",
        ),
        (
            ClickAttrs.tuple_widget(types=(str, int, float)),
            ("t", 1, -2.0),
            "main  cmd --p  t  1  -2.0",
        ),
        (
            ClickAttrs.nvalue_widget(type=(str, int)),
            [["a", 12], ["b", 11]],
            "main  cmd --p   a   12  --p   b   11 ",
        ),
        (
            (
                ClickAttrs.multi_value_widget(nargs=2),
                ["foo", "bar"],
                "main  cmd --p  foo  bar",
            )
        ),
        (
            ClickAttrs.multi_value_widget(nargs=2, default=["A", "B"]),
            ["A", "C"],
            "main  cmd --p  A  C",
        ),
        (
            ClickAttrs.multi_value_widget(nargs=2, default=["A", "B"]),
            [" ", " "],
            "main  cmd --p      ",
        ),
        (
            ClickAttrs.nvalue_widget(type=(click.types.File(), int)),
            [[".gitignore", 12], ["setup.py", -1]],
            "main  cmd --p   .gitignore   12  --p   setup.py   -1 ",
        ),
    ],
)
def test_command_with_ep_group(click_attrs: dict, value: t.Any, expected_output: str):
    param = click.Option(param_decls=["--p"], **click_attrs)
    cli = click.Group("cli")
    cmd = click.Command("cmd", params=[param])
    cli.add_command(cmd)
    control = clickqt.qtgui_from_click(cli)
    control.set_ep_or_path("main")
    control.set_is_ep(True)
    widget = control.widget_registry[prepare_execution(cmd, cli)][param.name]
    widget.set_value(value)

    control.construct_command_string()

    assert control.is_ep is True
    assert control.ep_or_path == "main"
    assert control.cmd == cli

    # Simulate clipboard behavior using QApplication.clipboard()
    clipboard = QApplication.clipboard()
    print(clipboard.text(QClipboard.Clipboard))
    assert clipboard.text(QClipboard.Clipboard) == expected_output


@pytest.mark.parametrize(
    ("click_attrs", "value", "expected_output"),
    [
        (ClickAttrs.intfield(), 12, "python example/example/main.py  cmd --p 12"),
        (
            ClickAttrs.textfield(),
            "test",
            "python example/example/main.py  cmd --p test",
        ),
        (ClickAttrs.realfield(), 0.8, "python example/example/main.py  cmd --p 0.8"),
        (
            ClickAttrs.passwordfield(),
            "abc",
            "python example/example/main.py  cmd --p abc",
        ),
        (ClickAttrs.checkbox(), True, "python example/example/main.py  cmd --p True"),
        (ClickAttrs.checkbox(), False, "python example/example/main.py  cmd --p False"),
        (
            ClickAttrs.intrange(maxval=2, clamp=True),
            5,
            "python example/example/main.py  cmd --p 2",
        ),
        (
            ClickAttrs.floatrange(maxval=2.05, clamp=True),
            5,
            "python example/example/main.py  cmd --p 2.05",
        ),
        (
            ClickAttrs.combobox(
                choices=["A", "B", "C"], case_sensitive=False, confirmation_prompt=True
            ),
            "B",
            "python example/example/main.py  cmd --p=B",
        ),
        (
            ClickAttrs.combobox(choices=["A", "B", "C"], case_sensitive=False),
            "B",
            "python example/example/main.py  cmd --p=B",
        ),
        (
            ClickAttrs.checkable_combobox(choices=["A", "B", "C"]),
            ["B", "C"],
            "python example/example/main.py  cmd --p=B --p=C",
        ),
        (
            ClickAttrs.checkable_combobox(choices=["A", "B", "C"]),
            ["A"],
            "python example/example/main.py  cmd --p=A",
        ),
        (
            ClickAttrs.checkable_combobox(choices=["A", "B", "C"]),
            ["A", "B", "C"],
            "python example/example/main.py  cmd --p=A --p=B --p=C",
        ),
        (
            ClickAttrs.tuple_widget(types=(str, int, float)),
            ("t", 1, -2.0),
            "python example/example/main.py  cmd --p  t  1  -2.0",
        ),
        (
            ClickAttrs.nvalue_widget(type=(str, int)),
            [["a", 12], ["b", 11]],
            "python example/example/main.py  cmd --p   a   12  --p   b   11 ",
        ),
        (
            (
                ClickAttrs.multi_value_widget(nargs=2),
                ["foo", "bar"],
                "python example/example/main.py  cmd --p  foo  bar",
            )
        ),
        (
            ClickAttrs.multi_value_widget(nargs=2, default=["A", "B"]),
            ["A", "C"],
            "python example/example/main.py  cmd --p  A  C",
        ),
        (
            ClickAttrs.multi_value_widget(nargs=2, default=["A", "B"]),
            [" ", " "],
            "python example/example/main.py  cmd --p      ",
        ),
        (
            ClickAttrs.nvalue_widget(type=(click.types.File(), int)),
            [[".gitignore", 12], ["setup.py", -1]],
            "python example/example/main.py  cmd --p   .gitignore   12  --p   setup.py   -1 ",
        ),
    ],
)
def test_construct_cmd_string_file_grouped(
    click_attrs: dict, value: t.Any, expected_output: str
):
    param = click.Option(param_decls=["--p"], **click_attrs)
    cli = click.Group("cli")
    cmd = click.Command("cmd", params=[param])
    cli.add_command(cmd)
    control = clickqt.qtgui_from_click(cli)
    control.set_ep_or_path("example/example/main.py")
    control.set_is_ep(False)
    widget = control.widget_registry[prepare_execution(cmd, cli)][param.name]
    widget.set_value(value)

    control.construct_command_string()

    assert control.is_ep is False
    assert control.ep_or_path == "example/example/main.py"
    assert control.cmd == cli

    # Simulate clipboard behavior using QApplication.clipboard()
    clipboard = QApplication.clipboard()
    print(clipboard.text(QClipboard.Clipboard))
    assert clipboard.text(QClipboard.Clipboard) == expected_output
